#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Desc   :

from pathlib import Path
from typing import Union

import numpy as np
import numpy.typing as npt
from gymnasium import spaces
from pydantic import ConfigDict, Field, field_validator

from metagpt.environment.base_env_space import (
    BaseEnvAction,
    BaseEnvActionType,
    BaseEnvObsParams,
    BaseEnvObsType,
)


class EnvActionType(BaseEnvActionType):
    NONE = 0  # no action to run, just get observation

    SYSTEM_BACK = 1
    SYSTEM_TAP = 2
    USER_INPUT = 3
    USER_LONGPRESS = 4
    USER_SWIPE = 5
    USER_SWIPE_TO = 6


class EnvAction(BaseEnvAction):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    action_type: int = Field(default=EnvActionType.NONE, description="action type")
    coord: npt.NDArray[np.int64] = Field(
        default_factory=lambda: np.zeros(2, dtype=np.int64), description="operation coordinate"
    )
    tgt_coord: npt.NDArray[np.int64] = Field(
        default_factory=lambda: np.zeros(2, dtype=np.int64), description="target operation coordinate"
    )
    input_txt: str = Field(default="", description="user input text")
    orient: str = Field(default="up", description="swipe orient")
    dist: str = Field(default="medium", description="swipe dist")

    @field_validator("coord", "tgt_coord", mode="before")
    @classmethod
    def check_coord(cls, coord) -> npt.NDArray[np.int64]:
        if not isinstance(coord, np.ndarray):
            return np.array(coord)


class EnvObsType(BaseEnvObsType):
    NONE = 0  # get whole observation from env

    GET_SCREENSHOT = 1
    GET_XML = 2


class EnvObsParams(BaseEnvObsParams):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    obs_type: int = Field(default=EnvObsType.NONE, description="observation type")
    ss_name: str = Field(default="", description="screenshot file name")
    xml_name: str = Field(default="", description="xml file name")
    local_save_dir: Union[str, Path] = Field(default="", description="local dir to save file")


EnvObsValType = str


def get_observation_space() -> spaces.Dict:
    space = spaces.Dict({"screenshot": spaces.Text(256), "xml": spaces.Text(256)})
    return space


def get_action_space(device_shape: tuple[int, int]) -> spaces.Dict:
    space = spaces.Dict(
        {
            "action_type": spaces.Discrete(len(EnvActionType)),
            "coord": spaces.Box(
                np.array([0, 0], dtype=np.int64), np.array([device_shape[0], device_shape[1]], dtype=np.int64)
            ),
            "tgt_coord": spaces.Box(
                np.array([0, 0], dtype=np.int64), np.array([device_shape[0], device_shape[1]], dtype=np.int64)
            ),
            "input_txt": spaces.Text(256),
            "orient": spaces.Text(16),
            "dist": spaces.Text(16),
        }
    )
    return space



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This file provides functionality to convert a local repository into a markdown representation.
"""
from __future__ import annotations

import mimetypes
from pathlib import Path

from gitignore_parser import parse_gitignore

from metagpt.logs import logger
from metagpt.utils.common import aread, awrite, get_markdown_codeblock_type, list_files
from metagpt.utils.tree import tree


async def repo_to_markdown(repo_path: str | Path, output: str | Path = None, gitignore: str | Path = None) -> str:
    """
    Convert a local repository into a markdown representation.

    This function takes a path to a local repository and generates a markdown representation of the repository structure,
    including directory trees and file listings.

    Args:
        repo_path (str | Path): The path to the local repository.
        output (str | Path, optional): The path to save the generated markdown file. Defaults to None.
        gitignore (str | Path, optional): The path to the .gitignore file. Defaults to None.

    Returns:
        str: The markdown representation of the repository.
    """
    repo_path = Path(repo_path)
    gitignore = Path(gitignore or Path(__file__).parent / "../../.gitignore").resolve()

    markdown = await _write_dir_tree(repo_path=repo_path, gitignore=gitignore)

    gitignore_rules = parse_gitignore(full_path=str(gitignore))
    markdown += await _write_files(repo_path=repo_path, gitignore_rules=gitignore_rules)

    if output:
        await awrite(filename=str(output), data=markdown, encoding="utf-8")
    return markdown


async def _write_dir_tree(repo_path: Path, gitignore: Path) -> str:
    try:
        content = tree(repo_path, gitignore, run_command=True)
    except Exception as e:
        logger.info(f"{e}, using safe mode.")
        content = tree(repo_path, gitignore, run_command=False)

    doc = f"## Directory Tree\n```text\n{content}\n```\n---\n\n"
    return doc


async def _write_files(repo_path, gitignore_rules) -> str:
    filenames = list_files(repo_path)
    markdown = ""
    for filename in filenames:
        if gitignore_rules(str(filename)):
            continue
        markdown += await _write_file(filename=filename, repo_path=repo_path)
    return markdown


async def _write_file(filename: Path, repo_path: Path) -> str:
    relative_path = filename.relative_to(repo_path)
    markdown = f"## {relative_path}\n"

    mime_type, _ = mimetypes.guess_type(filename.name)
    if "text/" not in mime_type:
        logger.info(f"Ignore content: {filename}")
        markdown += "<binary file>\n---\n\n"
        return markdown
    content = await aread(filename, encoding="utf-8")
    content = content.replace("```", "\\`\\`\\`").replace("---", "\\-\\-\\-")
    code_block_type = get_markdown_codeblock_type(filename.name)
    markdown += f"```{code_block_type}\n{content}\n```\n---\n\n"
    return markdown



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Desc   : dummy action to make every STRole can deal DummyMessage which is caused by DummyAction

from metagpt.actions import Action
from metagpt.schema import Message


class DummyAction(Action):
    async def run(self, *args, **kwargs):
        raise NotImplementedError


class DummyMessage(Message):
    """
    dummy message to pass to role and make them to have a execution every round
    """

    content: str = "dummy"
    cause_by: str = "DummyAction"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Time    : 2023/01/12 17:07
@Author  : garylin2099
@File    : tool_registry.py
"""
from __future__ import annotations

import inspect
import os
from collections import defaultdict
from pathlib import Path

import yaml
from pydantic import BaseModel

from metagpt.const import TOOL_SCHEMA_PATH
from metagpt.logs import logger
from metagpt.tools.tool_convert import (
    convert_code_to_tool_schema,
    convert_code_to_tool_schema_ast,
)
from metagpt.tools.tool_data_type import Tool, ToolSchema


class ToolRegistry(BaseModel):
    tools: dict = {}
    tools_by_tags: dict = defaultdict(dict)  # two-layer k-v, {tag: {tool_name: {...}, ...}, ...}

    def register_tool(
        self,
        tool_name: str,
        tool_path: str,
        schemas: dict = None,
        schema_path: str = "",
        tool_code: str = "",
        tags: list[str] = None,
        tool_source_object=None,  # can be any classes or functions
        include_functions: list[str] = None,
        verbose: bool = False,
    ):
        if self.has_tool(tool_name):
            return

        schema_path = schema_path or TOOL_SCHEMA_PATH / f"{tool_name}.yml"

        if not schemas:
            schemas = make_schema(tool_source_object, include_functions, schema_path)

        if not schemas:
            return

        schemas["tool_path"] = tool_path  # corresponding code file path of the tool
        try:
            ToolSchema(**schemas)  # validation
        except Exception:
            pass
            # logger.warning(
            #     f"{tool_name} schema not conforms to required format, but will be used anyway. Mismatch: {e}"
            # )
        tags = tags or []
        tool = Tool(name=tool_name, path=tool_path, schemas=schemas, code=tool_code, tags=tags)
        self.tools[tool_name] = tool
        for tag in tags:
            self.tools_by_tags[tag].update({tool_name: tool})
        if verbose:
            logger.info(f"{tool_name} registered")
            logger.info(f"schema made at {str(schema_path)}, can be used for checking")

    def has_tool(self, key: str) -> Tool:
        return key in self.tools

    def get_tool(self, key) -> Tool:
        return self.tools.get(key)

    def get_tools_by_tag(self, key) -> dict[str, Tool]:
        return self.tools_by_tags.get(key, {})

    def get_all_tools(self) -> dict[str, Tool]:
        return self.tools

    def has_tool_tag(self, key) -> bool:
        return key in self.tools_by_tags

    def get_tool_tags(self) -> list[str]:
        return list(self.tools_by_tags.keys())


# Registry instance
TOOL_REGISTRY = ToolRegistry()


def register_tool(tags: list[str] = None, schema_path: str = "", **kwargs):
    """register a tool to registry"""

    def decorator(cls):
        # Get the file path where the function / class is defined and the source code
        file_path = inspect.getfile(cls)
        if "metagpt" in file_path:
            # split to handle ../metagpt/metagpt/tools/... where only metapgt/tools/... is needed
            file_path = "metagpt" + file_path.split("metagpt")[-1]
        source_code = inspect.getsource(cls)

        TOOL_REGISTRY.register_tool(
            tool_name=cls.__name__,
            tool_path=file_path,
            schema_path=schema_path,
            tool_code=source_code,
            tags=tags,
            tool_source_object=cls,
            **kwargs,
        )
        return cls

    return decorator


def make_schema(tool_source_object, include, path):
    os.makedirs(os.path.dirname(path), exist_ok=True)  # Create the necessary directories
    try:
        schema = convert_code_to_tool_schema(tool_source_object, include=include)
        with open(path, "w", encoding="utf-8") as f:
            yaml.dump(schema, f, sort_keys=False)
    except Exception as e:
        schema = {}
        logger.error(f"Fail to make schema: {e}")

    return schema


def validate_tool_names(tools: list[str]) -> dict[str, Tool]:
    assert isinstance(tools, list), "tools must be a list of str"
    valid_tools = {}
    for key in tools:
        # one can define either tool names OR tool tags OR tool path, take union to get the whole set
        # if tool paths are provided, they will be registered on the fly
        if os.path.isdir(key) or os.path.isfile(key):
            valid_tools.update(register_tools_from_path(key))
        elif TOOL_REGISTRY.has_tool(key):
            valid_tools.update({key: TOOL_REGISTRY.get_tool(key)})
        elif TOOL_REGISTRY.has_tool_tag(key):
            valid_tools.update(TOOL_REGISTRY.get_tools_by_tag(key))
        else:
            logger.warning(f"invalid tool name or tool type name: {key}, skipped")
    return valid_tools


def register_tools_from_file(file_path) -> dict[str, Tool]:
    file_name = Path(file_path).name
    if not file_name.endswith(".py") or file_name == "setup.py" or file_name.startswith("test"):
        return {}
    registered_tools = {}
    code = Path(file_path).read_text(encoding="utf-8")
    tool_schemas = convert_code_to_tool_schema_ast(code)
    for name, schemas in tool_schemas.items():
        tool_code = schemas.pop("code", "")
        TOOL_REGISTRY.register_tool(
            tool_name=name,
            tool_path=file_path,
            schemas=schemas,
            tool_code=tool_code,
        )
        registered_tools.update({name: TOOL_REGISTRY.get_tool(name)})
    return registered_tools


def register_tools_from_path(path) -> dict[str, Tool]:
    tools_registered = {}
    if os.path.isfile(path):
        tools_registered.update(register_tools_from_file(path))
    elif os.path.isdir(path):
        for root, _, files in os.walk(path):
            for file in files:
                file_path = os.path.join(root, file)
                tools_registered.update(register_tools_from_file(file_path))
    return tools_registered



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Desc   : zhipuai LLM from https://open.bigmodel.cn/dev/api#sdk

from enum import Enum
from typing import Optional

from zhipuai.types.chat.chat_completion import Completion

from metagpt.configs.llm_config import LLMConfig, LLMType
from metagpt.const import USE_CONFIG_TIMEOUT
from metagpt.logs import log_llm_stream
from metagpt.provider.base_llm import BaseLLM
from metagpt.provider.llm_provider_registry import register_provider
from metagpt.provider.zhipuai.zhipu_model_api import ZhiPuModelAPI
from metagpt.utils.cost_manager import CostManager


class ZhiPuEvent(Enum):
    ADD = "add"
    ERROR = "error"
    INTERRUPTED = "interrupted"
    FINISH = "finish"


@register_provider(LLMType.ZHIPUAI)
class ZhiPuAILLM(BaseLLM):
    """
    Refs to `https://open.bigmodel.cn/dev/api#chatglm_turbo`
    From now, support glm-3-turbo、glm-4, and also system_prompt.
    """

    def __init__(self, config: LLMConfig):
        self.config = config
        self.__init_zhipuai()
        self.cost_manager: Optional[CostManager] = None

    def __init_zhipuai(self):
        assert self.config.api_key
        self.api_key = self.config.api_key
        self.model = self.config.model  # so far, it support glm-3-turbo、glm-4
        self.pricing_plan = self.config.pricing_plan or self.model
        self.llm = ZhiPuModelAPI(api_key=self.api_key)

    def _const_kwargs(self, messages: list[dict], stream: bool = False) -> dict:
        max_tokens = self.config.max_token if self.config.max_token > 0 else 1024
        temperature = self.config.temperature if self.config.temperature > 0.0 else 0.3
        kwargs = {
            "model": self.model,
            "max_tokens": max_tokens,
            "messages": messages,
            "stream": stream,
            "temperature": temperature,
        }
        return kwargs

    def completion(self, messages: list[dict], timeout=USE_CONFIG_TIMEOUT) -> dict:
        resp: Completion = self.llm.chat.completions.create(**self._const_kwargs(messages))
        usage = resp.usage.model_dump()
        self._update_costs(usage)
        return resp.model_dump()

    async def _achat_completion(self, messages: list[dict], timeout=USE_CONFIG_TIMEOUT) -> dict:
        resp = await self.llm.acreate(**self._const_kwargs(messages))
        usage = resp.get("usage", {})
        self._update_costs(usage)
        return resp

    async def acompletion(self, messages: list[dict], timeout=USE_CONFIG_TIMEOUT) -> dict:
        return await self._achat_completion(messages, timeout=self.get_timeout(timeout))

    async def _achat_completion_stream(self, messages: list[dict], timeout=USE_CONFIG_TIMEOUT) -> str:
        response = await self.llm.acreate_stream(**self._const_kwargs(messages, stream=True))
        collected_content = []
        usage = {}
        async for chunk in response.stream():
            finish_reason = chunk.get("choices")[0].get("finish_reason")
            if finish_reason == "stop":
                usage = chunk.get("usage", {})
            else:
                content = self.get_choice_delta_text(chunk)
                collected_content.append(content)
                log_llm_stream(content)

        log_llm_stream("\n")

        self._update_costs(usage)
        full_content = "".join(collected_content)
        return full_content



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Desc   : Integration Reflect Action

import re

from metagpt.ext.stanford_town.actions.st_action import STAction
from metagpt.logs import logger


# Run GPT Prompt Focal Point method
class AgentFocusPt(STAction):
    name: str = "AgentFocusPt"

    def _func_validate(self, llm_resp: str, prompt: str) -> bool:
        try:
            self._func_cleanup(llm_resp, prompt)
            return True
        except Exception:
            return False

    def _func_cleanup(self, llm_resp: str, prompt: str = "") -> str:
        try:
            """
            Cleanup handling has been completed for run_v2
            """
            return llm_resp
        except Exception as exp:
            logger.error(f"{self.cls_name} with error {exp}")

    def _func_fail_default_resp(self) -> str:
        pass

    async def run(self, role: "STRole", statements: str, n: int, test_input=None) -> str:
        def create_prompt_input(role: "STRole", statements, n, test_input=None):
            prompt_input = [statements, str(n)]
            return prompt_input

        prompt_input = create_prompt_input(role, statements, n)
        prompt = self.generate_prompt_with_tmpl_filename(prompt_input, "generate_focal_pt_v1.txt")

        example_output = '["What should Jane do for lunch", "Does Jane like strawberry", "Who is Jane"]'
        special_instruction = "Output must be a list of str."
        output = await self._run_gpt35(prompt, example_output, special_instruction)
        logger.info(f"Role: {role.name} Action: {self.cls_name} output: {output}")
        return output


# Run GPT Prompt Insight and Guidance
class AgentInsightAndGuidance(STAction):
    name: str = "AgentInsightAndGuidance"

    def _func_validate(self, llm_resp: str, prompt: str) -> bool:
        try:
            self._func_cleanup(llm_resp, prompt)
            return True
        except Exception:
            return False

    def _func_cleanup(self, llm_resp: str, prompt: str = "") -> dict:
        try:
            llm_resp = "1. " + llm_resp.strip()
            ret = dict()
            for i in llm_resp.split("\n"):
                row = " ".join(i.split(". ")[1:])
                if "(because of " not in row:
                    continue
                thought = row.split("(because of ")[0].strip()
                if ")" not in row.split("(because of ")[1]:
                    continue
                evi_raw = row.split("(because of ")[1].split(")")[0].strip()
                evi_raw = re.findall(r"\d+", evi_raw)
                evi_raw = [int(i.strip()) for i in evi_raw]
                ret[thought] = evi_raw
            return ret
        except Exception as exp:
            logger.error(f"{self.cls_name} with error {exp}")

    def _func_fail_default_resp(self, n: int) -> str:
        return ["I am hungry"] * n

    async def run(self, role: "STRole", statements: str, n: int, test_input=None) -> dict:
        def create_prompt_input(role, statements, n, test_input=None):
            prompt_input = [statements, str(n)]
            return prompt_input

        prompt_input = create_prompt_input(role, statements, n)
        prompt = self.generate_prompt_with_tmpl_filename(prompt_input, "insight_and_evidence_v1.txt")

        self.fail_default_resp = self._func_fail_default_resp(n)
        output = await self._run_gpt35_max_tokens(prompt, max_tokens=150)
        logger.info(f"Role: {role.name} Action: {self.cls_name} output: {output}")
        return output


# Run GPT Prompt Event Triple
class AgentEventTriple(STAction):
    name: str = "AgentEventTriple"

    def _func_validate(self, llm_resp: str, prompt: str) -> bool:
        try:
            llm_resp = self._func_cleanup(llm_resp, prompt="")
            if len(llm_resp) != 2:
                return False
        except Exception:
            return False
        return True

    def _func_cleanup(self, llm_resp: str, prompt: str = "") -> list:
        try:
            cr = llm_resp.strip()
            cr = [i.strip() for i in cr.split(")")[0].split(",")]
            if len(cr) != 2:
                return cr[-2:]
            return cr
        except Exception as exp:
            logger.error(f"{self.cls_name} with error {exp}")

    def _func_fail_default_resp(self) -> str:
        pass

    async def run(self, statements: str, role: "STRole", verbose=False) -> tuple:
        def create_prompt_input(statements, role):
            if "(" in statements:
                statements = statements.split("(")[-1].split(")")[0]
            prompt_input = [role.scratch.name, statements, role.scratch.name]
            return prompt_input

        prompt_input = create_prompt_input(statements, role)
        prompt = self.generate_prompt_with_tmpl_filename(prompt_input, "generate_event_triple_v1.txt")

        output = await self._run_gpt35_max_tokens(prompt, max_tokens=30)
        output = (role.scratch.name, output[0], output[1])
        logger.info(f"Role: {role.name} Action: {self.cls_name} output: {output}")
        return output


# Run GPT Prompt Event Poignancy
class AgentEventPoignancy(STAction):
    name: str = "AgentEventPoignancy"

    def _func_validate(self, llm_resp: str, prompt: str) -> bool:
        try:
            self._func_cleanup(llm_resp, prompt)
            return True
        except Exception:
            return False

    def _func_cleanup(self, llm_resp: str, prompt: str = "") -> int:
        try:
            llm_resp = int(llm_resp.strip())
            return llm_resp
        except Exception as exp:
            logger.error(f"{self.cls_name} with error {exp}")

    def _func_fail_default_resp(self) -> str:
        pass

    async def run(self, role: "STRole", statements: str, test_input=None, verbose=False) -> str:
        def create_prompt_input(role: "STRole", statements: str, test_input=None):
            prompt_input = [role.scratch.name, role.scratch.get_str_iss(), role.scratch.name, statements]
            return prompt_input

        prompt_input = create_prompt_input(role, statements)
        prompt = self.generate_prompt_with_tmpl_filename(prompt_input, "poignancy_event_v1.txt")

        example_output = "5"  # ########
        special_instruction = "The output should ONLY contain ONE integer value on the scale of 1 to 10."
        output = await self._run_gpt35(prompt, example_output, special_instruction)
        logger.info(f"Role: {role.name} Action: {self.cls_name} output: {output}")
        return output


# Run GPT Prompt Chat Poignancy
class AgentChatPoignancy(STAction):
    name: str = "AgentChatPoignancy"

    def _func_validate(self, llm_resp: str, prompt: str) -> bool:
        try:
            self._func_cleanup(llm_resp, prompt)
            return True
        except Exception:
            return False

    def _func_cleanup(self, llm_resp: str, prompt: str = "") -> int:
        try:
            llm_resp = int(llm_resp.strip())
            return llm_resp
        except Exception as exp:
            logger.error(f"{self.cls_name} with error {exp}")

    def _func_fail_default_resp(self) -> str:
        pass

    async def run(self, role: "STRole", statements: str, test_input=None, verbose=False) -> str:
        def create_prompt_input(role: "STRole", statements, test_input=None):
            prompt_input = [role.scratch.name, role.scratch.get_str_iss(), role.scratch.name, statements]
            return prompt_input

        prompt_input = create_prompt_input(role, statements)
        prompt = self.generate_prompt_with_tmpl_filename(prompt_input, "poignancy_chat_v1.txt")

        example_output = "5"  # ########
        special_instruction = "The output should ONLY contain ONE integer value on the scale of 1 to 10."
        output = await self._run_gpt35(prompt, example_output, special_instruction)
        logger.info(f"Role: {role.name} Action: {self.cls_name} output: {output}")
        return output


# Run GPT Prompt Planning Thought on Convo
class AgentPlanThoughtOnConvo(STAction):
    name: str = "AgentPlanThoughtOnConvo"

    def _func_validate(self, llm_resp: str, prompt: str) -> bool:
        try:
            self._func_cleanup(llm_resp, prompt)
            return True
        except Exception:
            return False

    def _func_cleanup(self, llm_resp: str, prompt: str = "") -> str:
        try:
            return llm_resp.split('"')[0].strip()
        except Exception as exp:
            logger.error(f"{self.cls_name} with error {exp}")

    def _func_fail_default_resp(self) -> str:
        pass

    async def run(self, role: "STRole", statements: str, test_input=None, verbose=False) -> str:
        def create_prompt_input(role, statements, test_input=None):
            prompt_input = [statements, role.scratch.name, role.scratch.name, role.scratch.name]
            return prompt_input

        prompt_input = create_prompt_input(role, statements)
        prompt = self.generate_prompt_with_tmpl_filename(prompt_input, "planning_thought_on_convo_v1.txt")

        output = await self._run_gpt35_max_tokens(prompt, max_tokens=50)
        logger.info(f"Role: {role.name} Action: {self.cls_name} output: {output}")
        return output


# Run GPT Prompt Memory on Convo
class AgentMemoryOnConvo(STAction):
    name: str = "AgentMemoryOnConvo"

    def _func_validate(self, llm_resp: str, prompt: str) -> bool:
        try:
            self._func_cleanup(llm_resp, prompt)
            return True
        except Exception:
            return False

    def _func_cleanup(self, llm_resp: str, prompt: str = "") -> str:
        try:
            return llm_resp.split('"')[0].strip()
        except Exception as exp:
            logger.error(f"{self.cls_name} with error {exp}")

    def _func_fail_default_resp(self) -> str:
        pass

    async def run(self, role: "STRole", statements: str, test_input=None, verbose=False) -> str:
        def create_prompt_input(role, statements, test_input=None):
            prompt_input = [statements, role.scratch.name, role.scratch.name, role.scratch.name]
            return prompt_input

        prompt_input = create_prompt_input(role, statements)
        prompt = self.generate_prompt_with_tmpl_filename(prompt_input, "memo_on_convo_v1.txt")
        example_output = "Jane Doe was interesting to talk to."
        special_instruction = (
            "The output should ONLY contain a string that summarizes anything interesting "
            "that the agent may have noticed"
        )
        output = await self._run_gpt35(prompt, example_output, special_instruction)
        logger.info(f"Role: {role.name} Action: {self.cls_name} output: {output}")
        return output



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Time    : 2023/12/19
@Author  : mashenquan
@File    : graph_repository.py
@Desc    : Superclass for graph repository. This script defines a superclass for a graph repository, providing a
    foundation for specific implementations.

"""

from abc import ABC, abstractmethod
from collections import defaultdict
from pathlib import Path
from typing import List

from pydantic import BaseModel

from metagpt.repo_parser import DotClassInfo, DotClassRelationship, RepoFileInfo
from metagpt.utils.common import concat_namespace, split_namespace


class GraphKeyword:
    """Basic words for a Graph database.

    This class defines a set of basic words commonly used in the context of a Graph database.
    """

    IS = "is"
    OF = "Of"
    ON = "On"
    CLASS = "class"
    FUNCTION = "function"
    HAS_FUNCTION = "has_function"
    SOURCE_CODE = "source_code"
    NULL = "<null>"
    GLOBAL_VARIABLE = "global_variable"
    CLASS_METHOD = "class_method"
    CLASS_PROPERTY = "class_property"
    HAS_CLASS_METHOD = "has_class_method"
    HAS_CLASS_PROPERTY = "has_class_property"
    HAS_CLASS = "has_class"
    HAS_DETAIL = "has_detail"
    HAS_PAGE_INFO = "has_page_info"
    HAS_CLASS_VIEW = "has_class_view"
    HAS_SEQUENCE_VIEW = "has_sequence_view"
    HAS_SEQUENCE_VIEW_VER = "has_sequence_view_ver"
    HAS_CLASS_USE_CASE = "has_class_use_case"
    IS_COMPOSITE_OF = "is_composite_of"
    IS_AGGREGATE_OF = "is_aggregate_of"
    HAS_PARTICIPANT = "has_participant"


class SPO(BaseModel):
    """Graph repository record type.

    This class represents a record in a graph repository with three components:
    - Subject: The subject of the triple.
    - Predicate: The predicate describing the relationship between the subject and the object.
    - Object: The object of the triple.

    Attributes:
        subject (str): The subject of the triple.
        predicate (str): The predicate describing the relationship.
        object_ (str): The object of the triple.

    Example:
        spo_record = SPO(subject="Node1", predicate="connects_to", object_="Node2")
        # Represents a triple: Node1 connects_to Node2
    """

    subject: str
    predicate: str
    object_: str


class GraphRepository(ABC):
    """Abstract base class for a Graph Repository.

    This class defines the interface for a graph repository, providing methods for inserting, selecting,
    deleting, and saving graph data. Concrete implementations of this class must provide functionality
    for these operations.
    """

    def __init__(self, name: str, **kwargs):
        self._repo_name = name
        self._kwargs = kwargs

    @abstractmethod
    async def insert(self, subject: str, predicate: str, object_: str):
        """Insert a new triple into the graph repository.

        Args:
            subject (str): The subject of the triple.
            predicate (str): The predicate describing the relationship.
            object_ (str): The object of the triple.

        Example:
            await my_repository.insert(subject="Node1", predicate="connects_to", object_="Node2")
            # Inserts a triple: Node1 connects_to Node2 into the graph repository.
        """
        pass

    @abstractmethod
    async def select(self, subject: str = None, predicate: str = None, object_: str = None) -> List[SPO]:
        """Retrieve triples from the graph repository based on specified criteria.

        Args:
            subject (str, optional): The subject of the triple to filter by.
            predicate (str, optional): The predicate describing the relationship to filter by.
            object_ (str, optional): The object of the triple to filter by.

        Returns:
            List[SPO]: A list of SPO objects representing the selected triples.

        Example:
            selected_triples = await my_repository.select(subject="Node1", predicate="connects_to")
            # Retrieves triples where Node1 is the subject and the predicate is 'connects_to'.
        """
        pass

    @abstractmethod
    async def delete(self, subject: str = None, predicate: str = None, object_: str = None) -> int:
        """Delete triples from the graph repository based on specified criteria.

        Args:
            subject (str, optional): The subject of the triple to filter by.
            predicate (str, optional): The predicate describing the relationship to filter by.
            object_ (str, optional): The object of the triple to filter by.

        Returns:
            int: The number of triples deleted from the repository.

        Example:
            deleted_count = await my_repository.delete(subject="Node1", predicate="connects_to")
            # Deletes triples where Node1 is the subject and the predicate is 'connects_to'.
        """
        pass

    @abstractmethod
    async def save(self):
        """Save any changes made to the graph repository.

        Example:
            await my_repository.save()
            # Persists any changes made to the graph repository.
        """
        pass

    @property
    def name(self) -> str:
        """Get the name of the graph repository."""
        return self._repo_name

    @staticmethod
    async def update_graph_db_with_file_info(graph_db: "GraphRepository", file_info: RepoFileInfo):
        """Insert information of RepoFileInfo into the specified graph repository.

        This function updates the provided graph repository with information from the given RepoFileInfo object.
        The function inserts triples related to various dimensions such as file type, class, class method, function,
        global variable, and page info.

        Triple Patterns:
        - (?, is, [file type])
        - (?, has class, ?)
        - (?, is, [class])
        - (?, has class method, ?)
        - (?, has function, ?)
        - (?, is, [function])
        - (?, is, global variable)
        - (?, has page info, ?)

        Args:
            graph_db (GraphRepository): The graph repository object to be updated.
            file_info (RepoFileInfo): The RepoFileInfo object containing information to be inserted.

        Example:
            await update_graph_db_with_file_info(my_graph_repo, my_file_info)
            # Updates 'my_graph_repo' with information from 'my_file_info'.
        """
        await graph_db.insert(subject=file_info.file, predicate=GraphKeyword.IS, object_=GraphKeyword.SOURCE_CODE)
        file_types = {".py": "python", ".js": "javascript"}
        file_type = file_types.get(Path(file_info.file).suffix, GraphKeyword.NULL)
        await graph_db.insert(subject=file_info.file, predicate=GraphKeyword.IS, object_=file_type)
        for c in file_info.classes:
            class_name = c.get("name", "")
            # file -> class
            await graph_db.insert(
                subject=file_info.file,
                predicate=GraphKeyword.HAS_CLASS,
                object_=concat_namespace(file_info.file, class_name),
            )
            # class detail
            await graph_db.insert(
                subject=concat_namespace(file_info.file, class_name),
                predicate=GraphKeyword.IS,
                object_=GraphKeyword.CLASS,
            )
            methods = c.get("methods", [])
            for fn in methods:
                await graph_db.insert(
                    subject=concat_namespace(file_info.file, class_name),
                    predicate=GraphKeyword.HAS_CLASS_METHOD,
                    object_=concat_namespace(file_info.file, class_name, fn),
                )
                await graph_db.insert(
                    subject=concat_namespace(file_info.file, class_name, fn),
                    predicate=GraphKeyword.IS,
                    object_=GraphKeyword.CLASS_METHOD,
                )
        for f in file_info.functions:
            # file -> function
            await graph_db.insert(
                subject=file_info.file, predicate=GraphKeyword.HAS_FUNCTION, object_=concat_namespace(file_info.file, f)
            )
            # function detail
            await graph_db.insert(
                subject=concat_namespace(file_info.file, f), predicate=GraphKeyword.IS, object_=GraphKeyword.FUNCTION
            )
        for g in file_info.globals:
            await graph_db.insert(
                subject=concat_namespace(file_info.file, g),
                predicate=GraphKeyword.IS,
                object_=GraphKeyword.GLOBAL_VARIABLE,
            )
        for code_block in file_info.page_info:
            if code_block.tokens:
                await graph_db.insert(
                    subject=concat_namespace(file_info.file, *code_block.tokens),
                    predicate=GraphKeyword.HAS_PAGE_INFO,
                    object_=code_block.model_dump_json(),
                )
            for k, v in code_block.properties.items():
                await graph_db.insert(
                    subject=concat_namespace(file_info.file, k, v),
                    predicate=GraphKeyword.HAS_PAGE_INFO,
                    object_=code_block.model_dump_json(),
                )

    @staticmethod
    async def update_graph_db_with_class_views(graph_db: "GraphRepository", class_views: List[DotClassInfo]):
        """Insert dot format class information into the specified graph repository.

        This function updates the provided graph repository with class information from the given list of DotClassInfo objects.
        The function inserts triples related to various aspects of class views, including source code, file type, class,
        class property, class detail, method, composition, and aggregation.

        Triple Patterns:
        - (?, is, source code)
        - (?, is, file type)
        - (?, has class, ?)
        - (?, is, class)
        - (?, has class property, ?)
        - (?, is, class property)
        - (?, has detail, ?)
        - (?, has method, ?)
        - (?, is composite of, ?)
        - (?, is aggregate of, ?)

        Args:
            graph_db (GraphRepository): The graph repository object to be updated.
            class_views (List[DotClassInfo]): List of DotClassInfo objects containing class information to be inserted.


        Example:
            await update_graph_db_with_class_views(my_graph_repo, [class_info1, class_info2])
            # Updates 'my_graph_repo' with class information from the provided list of DotClassInfo objects.
        """
        for c in class_views:
            filename, _ = c.package.split(":", 1)
            await graph_db.insert(subject=filename, predicate=GraphKeyword.IS, object_=GraphKeyword.SOURCE_CODE)
            file_types = {".py": "python", ".js": "javascript"}
            file_type = file_types.get(Path(filename).suffix, GraphKeyword.NULL)
            await graph_db.insert(subject=filename, predicate=GraphKeyword.IS, object_=file_type)
            await graph_db.insert(subject=filename, predicate=GraphKeyword.HAS_CLASS, object_=c.package)
            await graph_db.insert(
                subject=c.package,
                predicate=GraphKeyword.IS,
                object_=GraphKeyword.CLASS,
            )
            await graph_db.insert(subject=c.package, predicate=GraphKeyword.HAS_DETAIL, object_=c.model_dump_json())
            for vn, vt in c.attributes.items():
                # class -> property
                await graph_db.insert(
                    subject=c.package,
                    predicate=GraphKeyword.HAS_CLASS_PROPERTY,
                    object_=concat_namespace(c.package, vn),
                )
                # property detail
                await graph_db.insert(
                    subject=concat_namespace(c.package, vn),
                    predicate=GraphKeyword.IS,
                    object_=GraphKeyword.CLASS_PROPERTY,
                )
                await graph_db.insert(
                    subject=concat_namespace(c.package, vn),
                    predicate=GraphKeyword.HAS_DETAIL,
                    object_=vt.model_dump_json(),
                )
            for fn, ft in c.methods.items():
                # class -> function
                await graph_db.insert(
                    subject=c.package,
                    predicate=GraphKeyword.HAS_CLASS_METHOD,
                    object_=concat_namespace(c.package, fn),
                )
                # function detail
                await graph_db.insert(
                    subject=concat_namespace(c.package, fn),
                    predicate=GraphKeyword.IS,
                    object_=GraphKeyword.CLASS_METHOD,
                )
                await graph_db.insert(
                    subject=concat_namespace(c.package, fn),
                    predicate=GraphKeyword.HAS_DETAIL,
                    object_=ft.model_dump_json(),
                )
            for i in c.compositions:
                await graph_db.insert(
                    subject=c.package, predicate=GraphKeyword.IS_COMPOSITE_OF, object_=concat_namespace("?", i)
                )
            for i in c.aggregations:
                await graph_db.insert(
                    subject=c.package, predicate=GraphKeyword.IS_AGGREGATE_OF, object_=concat_namespace("?", i)
                )

    @staticmethod
    async def update_graph_db_with_class_relationship_views(
        graph_db: "GraphRepository", relationship_views: List[DotClassRelationship]
    ):
        """Insert class relationships and labels into the specified graph repository.

        This function updates the provided graph repository with class relationship information from the given list
        of DotClassRelationship objects. The function inserts triples representing relationships and labels between
        classes.

        Triple Patterns:
        - (?, is relationship of, ?)
        - (?, is relationship on, ?)

        Args:
            graph_db (GraphRepository): The graph repository object to be updated.
            relationship_views (List[DotClassRelationship]): List of DotClassRelationship objects containing
            class relationship information to be inserted.

        Example:
            await update_graph_db_with_class_relationship_views(my_graph_repo, [relationship1, relationship2])
            # Updates 'my_graph_repo' with class relationship information from the provided list of DotClassRelationship objects.

        """
        for r in relationship_views:
            await graph_db.insert(
                subject=r.src, predicate=GraphKeyword.IS + r.relationship + GraphKeyword.OF, object_=r.dest
            )
            if not r.label:
                continue
            await graph_db.insert(
                subject=r.src,
                predicate=GraphKeyword.IS + r.relationship + GraphKeyword.ON,
                object_=concat_namespace(r.dest, r.label),
            )

    @staticmethod
    async def rebuild_composition_relationship(graph_db: "GraphRepository"):
        """Append namespace-prefixed information to relationship SPO (Subject-Predicate-Object) objects in the graph
            repository.

        This function updates the provided graph repository by appending namespace-prefixed information to existing
        relationship SPO objects.

        Args:
            graph_db (GraphRepository): The graph repository object to be updated.
        """
        classes = await graph_db.select(predicate=GraphKeyword.IS, object_=GraphKeyword.CLASS)
        mapping = defaultdict(list)
        for c in classes:
            name = split_namespace(c.subject)[-1]
            mapping[name].append(c.subject)

        rows = await graph_db.select(predicate=GraphKeyword.IS_COMPOSITE_OF)
        for r in rows:
            ns, class_ = split_namespace(r.object_)
            if ns != "?":
                continue
            val = mapping[class_]
            if len(val) != 1:
                continue
            ns_name = val[0]
            await graph_db.delete(subject=r.subject, predicate=r.predicate, object_=r.object_)
            await graph_db.insert(subject=r.subject, predicate=r.predicate, object_=ns_name)



from enum import Enum
from typing import Optional

from pydantic import field_validator

from metagpt.utils.yaml_model import YamlModel


class EmbeddingType(Enum):
    OPENAI = "openai"
    AZURE = "azure"
    GEMINI = "gemini"
    OLLAMA = "ollama"


class EmbeddingConfig(YamlModel):
    """Config for Embedding.

    Examples:
    ---------
    api_type: "openai"
    api_key: "YOU_API_KEY"
    dimensions: "YOUR_MODEL_DIMENSIONS"

    api_type: "azure"
    api_key: "YOU_API_KEY"
    base_url: "YOU_BASE_URL"
    api_version: "YOU_API_VERSION"
    dimensions: "YOUR_MODEL_DIMENSIONS"

    api_type: "gemini"
    api_key: "YOU_API_KEY"

    api_type: "ollama"
    base_url: "YOU_BASE_URL"
    model: "YOU_MODEL"
    dimensions: "YOUR_MODEL_DIMENSIONS"
    """

    api_type: Optional[EmbeddingType] = None
    api_key: Optional[str] = None
    base_url: Optional[str] = None
    api_version: Optional[str] = None

    model: Optional[str] = None
    embed_batch_size: Optional[int] = None
    dimensions: Optional[int] = None  # output dimension of embedding model

    @field_validator("api_type", mode="before")
    @classmethod
    def check_api_type(cls, v):
        if v == "":
            return None
        return v



import asyncio
from fileinput import FileInput
from pathlib import Path
from typing import List, Optional, Union

from llama_index.core import Document
from llama_index.core.async_utils import run_jobs
from llama_index.core.readers.base import BaseReader

from metagpt.logs import logger
from metagpt.rag.schema import OmniParseOptions, OmniParseType, ParseResultType
from metagpt.utils.async_helper import NestAsyncio
from metagpt.utils.omniparse_client import OmniParseClient


class OmniParse(BaseReader):
    """OmniParse"""

    def __init__(
        self, api_key: str = None, base_url: str = "http://localhost:8000", parse_options: OmniParseOptions = None
    ):
        """
        Args:
            api_key: Default None, can be used for authentication later.
            base_url: OmniParse Base URL for the API.
            parse_options: Optional settings for OmniParse. Default is OmniParseOptions with default values.
        """
        self.parse_options = parse_options or OmniParseOptions()
        self.omniparse_client = OmniParseClient(api_key, base_url, max_timeout=self.parse_options.max_timeout)

    @property
    def parse_type(self):
        return self.parse_options.parse_type

    @property
    def result_type(self):
        return self.parse_options.result_type

    @parse_type.setter
    def parse_type(self, parse_type: Union[str, OmniParseType]):
        if isinstance(parse_type, str):
            parse_type = OmniParseType(parse_type)
        self.parse_options.parse_type = parse_type

    @result_type.setter
    def result_type(self, result_type: Union[str, ParseResultType]):
        if isinstance(result_type, str):
            result_type = ParseResultType(result_type)
        self.parse_options.result_type = result_type

    async def _aload_data(
        self,
        file_path: Union[str, bytes, Path],
        extra_info: Optional[dict] = None,
    ) -> List[Document]:
        """
        Load data from the input file_path.

        Args:
            file_path: File path or file byte data.
            extra_info: Optional dictionary containing additional information.

        Returns:
            List[Document]
        """
        try:
            if self.parse_type == OmniParseType.PDF:
                # pdf parse
                parsed_result = await self.omniparse_client.parse_pdf(file_path)
            else:
                # other parse use omniparse_client.parse_document
                # For compatible byte data, additional filename is required
                extra_info = extra_info or {}
                filename = extra_info.get("filename")
                parsed_result = await self.omniparse_client.parse_document(file_path, bytes_filename=filename)

            # Get the specified structured data based on result_type
            content = getattr(parsed_result, self.result_type)
            docs = [
                Document(
                    text=content,
                    metadata=extra_info or {},
                )
            ]
        except Exception as e:
            logger.error(f"OMNI Parse Error: {e}")
            docs = []

        return docs

    async def aload_data(
        self,
        file_path: Union[List[FileInput], FileInput],
        extra_info: Optional[dict] = None,
    ) -> List[Document]:
        """
        Load data from the input file_path.

        Args:
            file_path: File path or file byte data.
            extra_info: Optional dictionary containing additional information.

        Notes:
            This method ultimately calls _aload_data for processing.

        Returns:
            List[Document]
        """
        docs = []
        if isinstance(file_path, (str, bytes, Path)):
            # Processing single file
            docs = await self._aload_data(file_path, extra_info)
        elif isinstance(file_path, list):
            # Concurrently process multiple files
            parse_jobs = [self._aload_data(file_item, extra_info) for file_item in file_path]
            doc_ret_list = await run_jobs(jobs=parse_jobs, workers=self.parse_options.num_workers)
            docs = [doc for docs in doc_ret_list for doc in docs]
        return docs

    def load_data(
        self,
        file_path: Union[List[FileInput], FileInput],
        extra_info: Optional[dict] = None,
    ) -> List[Document]:
        """
        Load data from the input file_path.

        Args:
            file_path: File path or file byte data.
            extra_info: Optional dictionary containing additional information.

        Notes:
            This method ultimately calls aload_data for processing.

        Returns:
            List[Document]
        """
        NestAsyncio.apply_once()  # Ensure compatibility with nested async calls
        return asyncio.run(self.aload_data(file_path, extra_info))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Desc   : the ActionNode to parse screenshot

from metagpt.actions.action_node import ActionNode

OBSERVATION = ActionNode(
    key="Observation", expected_type=str, instruction="Describe what you observe in the image", example=""
)

THOUGHT = ActionNode(
    key="Thought",
    expected_type=str,
    instruction="To complete the given task, what is the next step I should do",
    example="",
)

ACTION = ActionNode(
    key="Action",
    expected_type=str,
    instruction="The function call with the correct parameters to proceed with the task. If you believe the task is "
    "completed or there is nothing to be done, you should output FINISH. You cannot output anything else "
    "except a function call or FINISH in this field.",
    example="",
)

SUMMARY = ActionNode(
    key="Summary",
    expected_type=str,
    instruction="Summarize your past actions along with your latest action in one or two sentences. Do not include "
    "the numeric tag in your summary",
    example="",
)

SUMMARY_GRID = ActionNode(
    key="Summary",
    expected_type=str,
    instruction="Summarize your past actions along with your latest action in one or two sentences. Do not include "
    "the grid area number in your summary",
    example="",
)

NODES = [OBSERVATION, THOUGHT, ACTION, SUMMARY]

NODES_GRID = [OBSERVATION, THOUGHT, ACTION, SUMMARY_GRID]

SCREENSHOT_PARSE_NODE = ActionNode.from_children("ScreenshotParse", NODES)
SCREENSHOT_PARSE_GRID_NODE = ActionNode.from_children("ScreenshotParseGrid", NODES_GRID)



import json
from abc import ABC, abstractmethod


class BaseBedrockProvider(ABC):
    # to handle different generation kwargs
    max_tokens_field_name = "max_tokens"

    @abstractmethod
    def _get_completion_from_dict(self, rsp_dict: dict) -> str:
        ...

    def get_request_body(self, messages: list[dict], const_kwargs, *args, **kwargs) -> str:
        body = json.dumps({"prompt": self.messages_to_prompt(messages), **const_kwargs})
        return body

    def get_choice_text(self, response_body: dict) -> str:
        completions = self._get_completion_from_dict(response_body)
        return completions

    def get_choice_text_from_stream(self, event) -> str:
        rsp_dict = json.loads(event["chunk"]["bytes"])
        completions = self._get_completion_from_dict(rsp_dict)
        return completions

    def messages_to_prompt(self, messages: list[dict]) -> str:
        """[{"role": "user", "content": msg}] to user: <msg> etc."""
        return "\n".join([f"{msg['role']}: {msg['content']}" for msg in messages])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Time    : 2024/1/4
@Author  : mashenquan
@File    : rebuild_sequence_view.py
@Desc    : Reconstruct sequence view information through reverse engineering.
    Implement RFC197, https://deepwisdom.feishu.cn/wiki/VyK0wfq56ivuvjklMKJcmHQknGt
"""
from __future__ import annotations

import re
from datetime import datetime
from pathlib import Path
from typing import List, Optional, Set

from pydantic import BaseModel
from tenacity import retry, stop_after_attempt, wait_random_exponential

from metagpt.actions import Action
from metagpt.config2 import config
from metagpt.const import GRAPH_REPO_FILE_REPO
from metagpt.logs import logger
from metagpt.repo_parser import CodeBlockInfo, DotClassInfo
from metagpt.schema import UMLClassView
from metagpt.utils.common import (
    add_affix,
    aread,
    auto_namespace,
    concat_namespace,
    general_after_log,
    list_files,
    parse_json_code_block,
    read_file_block,
    split_namespace,
)
from metagpt.utils.di_graph_repository import DiGraphRepository
from metagpt.utils.graph_repository import SPO, GraphKeyword, GraphRepository


class ReverseUseCase(BaseModel):
    """
    Represents a reverse engineered use case.

    Attributes:
        description (str): A description of the reverse use case.
        inputs (List[str]): List of inputs for the reverse use case.
        outputs (List[str]): List of outputs for the reverse use case.
        actors (List[str]): List of actors involved in the reverse use case.
        steps (List[str]): List of steps for the reverse use case.
        reason (str): The reason behind the reverse use case.
    """

    description: str
    inputs: List[str]
    outputs: List[str]
    actors: List[str]
    steps: List[str]
    reason: str


class ReverseUseCaseDetails(BaseModel):
    """
    Represents details of a reverse engineered use case.

    Attributes:
        description (str): A description of the reverse use case details.
        use_cases (List[ReverseUseCase]): List of reverse use cases.
        relationship (List[str]): List of relationships associated with the reverse use case details.
    """

    description: str
    use_cases: List[ReverseUseCase]
    relationship: List[str]


class RebuildSequenceView(Action):
    """
    Represents an action to reconstruct sequence view through reverse engineering.

    Attributes:
        graph_db (Optional[GraphRepository]): An optional instance of GraphRepository for graph database operations.
    """

    graph_db: Optional[GraphRepository] = None

    async def run(self, with_messages=None, format=config.prompt_schema):
        """
        Implementation of `Action`'s `run` method.

        Args:
            with_messages (Optional[Type]): An optional argument specifying messages to react to.
            format (str): The format for the prompt schema.
        """
        graph_repo_pathname = self.context.git_repo.workdir / GRAPH_REPO_FILE_REPO / self.context.git_repo.workdir.name
        self.graph_db = await DiGraphRepository.load_from(str(graph_repo_pathname.with_suffix(".json")))
        if not self.i_context:
            entries = await self._search_main_entry()
        else:
            entries = [SPO(subject=self.i_context, predicate="", object_="")]
        for entry in entries:
            await self._rebuild_main_sequence_view(entry)
            while await self._merge_sequence_view(entry):
                pass
        await self.graph_db.save()

    @retry(
        wait=wait_random_exponential(min=1, max=20),
        stop=stop_after_attempt(6),
        after=general_after_log(logger),
    )
    async def _rebuild_main_sequence_view(self, entry: SPO):
        """
        Reconstruct the sequence diagram for the __main__ entry of the source code through reverse engineering.

        Args:
            entry (SPO): The SPO (Subject, Predicate, Object) object in the graph database that is related to the
                subject `__name__:__main__`.
        """
        filename = entry.subject.split(":", 1)[0]
        rows = await self.graph_db.select(predicate=GraphKeyword.IS, object_=GraphKeyword.CLASS)
        classes = []
        prefix = filename + ":"
        for r in rows:
            if prefix in r.subject:
                classes.append(r)
                await self._rebuild_use_case(r.subject)
        participants = await self._search_participants(split_namespace(entry.subject)[0])
        class_details = []
        class_views = []
        for c in classes:
            detail = await self._get_class_detail(c.subject)
            if not detail:
                continue
            class_details.append(detail)
            view = await self._get_uml_class_view(c.subject)
            if view:
                class_views.append(view)

            actors = await self._get_participants(c.subject)
            participants.update(set(actors))

        use_case_blocks = []
        for c in classes:
            use_cases = await self._get_class_use_cases(c.subject)
            use_case_blocks.append(use_cases)
        prompt_blocks = ["## Use Cases\n" + "\n".join(use_case_blocks)]
        block = "## Participants\n"
        for p in participants:
            block += f"- {p}\n"
        prompt_blocks.append(block)
        block = "## Mermaid Class Views\n```mermaid\n"
        block += "\n\n".join([c.get_mermaid() for c in class_views])
        block += "\n```\n"
        prompt_blocks.append(block)
        block = "## Source Code\n```python\n"
        block += await self._get_source_code(filename)
        block += "\n```\n"
        prompt_blocks.append(block)
        prompt = "\n---\n".join(prompt_blocks)

        rsp = await self.llm.aask(
            msg=prompt,
            system_msgs=[
                "You are a python code to Mermaid Sequence Diagram translator in function detail.",
                "Translate the given markdown text to a Mermaid Sequence Diagram.",
                "Return the merged Mermaid sequence diagram in a markdown code block format.",
            ],
            stream=False,
        )
        sequence_view = rsp.removeprefix("```mermaid").removesuffix("```")
        rows = await self.graph_db.select(subject=entry.subject, predicate=GraphKeyword.HAS_SEQUENCE_VIEW)
        for r in rows:
            if r.predicate == GraphKeyword.HAS_SEQUENCE_VIEW:
                await self.graph_db.delete(subject=r.subject, predicate=r.predicate, object_=r.object_)
        await self.graph_db.insert(
            subject=entry.subject, predicate=GraphKeyword.HAS_SEQUENCE_VIEW, object_=sequence_view
        )
        await self.graph_db.insert(
            subject=entry.subject,
            predicate=GraphKeyword.HAS_SEQUENCE_VIEW_VER,
            object_=concat_namespace(datetime.now().strftime("%Y%m%d%H%M%S%f")[:-3], add_affix(sequence_view)),
        )
        for c in classes:
            await self.graph_db.insert(
                subject=entry.subject, predicate=GraphKeyword.HAS_PARTICIPANT, object_=auto_namespace(c.subject)
            )
        await self._save_sequence_view(subject=entry.subject, content=sequence_view)

    async def _merge_sequence_view(self, entry: SPO) -> bool:
        """
        Augments additional information to the provided SPO (Subject, Predicate, Object) entry in the sequence diagram.

        Args:
            entry (SPO): The SPO object representing the relationship in the graph database.

        Returns:
            bool: True if additional information has been augmented, otherwise False.
        """
        new_participant = await self._search_new_participant(entry)
        if not new_participant:
            return False

        await self._merge_participant(entry, new_participant)
        return True

    async def _search_main_entry(self) -> List:
        """
        Asynchronously searches for the SPO object that is related to `__name__:__main__`.

        Returns:
            List: A list containing information about the main entry in the sequence diagram.
        """
        rows = await self.graph_db.select(predicate=GraphKeyword.HAS_PAGE_INFO)
        tag = "__name__:__main__"
        entries = []
        for r in rows:
            if tag in r.subject or tag in r.object_:
                entries.append(r)
        return entries

    @retry(
        wait=wait_random_exponential(min=1, max=20),
        stop=stop_after_attempt(6),
        after=general_after_log(logger),
    )
    async def _rebuild_use_case(self, ns_class_name: str):
        """
        Asynchronously reconstructs the use case for the provided namespace-prefixed class name.

        Args:
            ns_class_name (str): The namespace-prefixed class name for which the use case is to be reconstructed.
        """
        rows = await self.graph_db.select(subject=ns_class_name, predicate=GraphKeyword.HAS_CLASS_USE_CASE)
        if rows:
            return

        detail = await self._get_class_detail(ns_class_name)
        if not detail:
            return
        participants = set()
        participants.update(set(detail.compositions))
        participants.update(set(detail.aggregations))
        class_view = await self._get_uml_class_view(ns_class_name)
        source_code = await self._get_source_code(ns_class_name)

        # prompt_blocks = [
        #     "## Instruction\n"
        #     "You are a python code to UML 2.0 Use Case translator.\n"
        #     'The generated UML 2.0 Use Case must include the roles or entities listed in "Participants".\n'
        #     "The functional descriptions of Actors and Use Cases in the generated UML 2.0 Use Case must not "
        #     'conflict with the information in "Mermaid Class Views".\n'
        #     'The section under `if __name__ == "__main__":` of "Source Code" contains information about external '
        #     "system interactions with the internal system.\n"
        # ]
        prompt_blocks = []
        block = "## Participants\n"
        for p in participants:
            block += f"- {p}\n"
        prompt_blocks.append(block)
        block = "## Mermaid Class Views\n```mermaid\n"
        block += class_view.get_mermaid()
        block += "\n```\n"
        prompt_blocks.append(block)
        block = "## Source Code\n```python\n"
        block += source_code
        block += "\n```\n"
        prompt_blocks.append(block)
        prompt = "\n---\n".join(prompt_blocks)

        rsp = await self.llm.aask(
            msg=prompt,
            system_msgs=[
                "You are a python code to UML 2.0 Use Case translator.",
                'The generated UML 2.0 Use Case must include the roles or entities listed in "Participants".',
                "The functional descriptions of Actors and Use Cases in the generated UML 2.0 Use Case must not "
                'conflict with the information in "Mermaid Class Views".',
                'The section under `if __name__ == "__main__":` of "Source Code" contains information about external '
                "system interactions with the internal system.",
                "Return a markdown JSON object with:\n"
                '- a "description" key to explain what the whole source code want to do;\n'
                '- a "use_cases" key list all use cases, each use case in the list should including a `description` '
                "key describes about what the use case to do, a `inputs` key lists the input names of the use case "
                "from external sources, a `outputs` key lists the output names of the use case to external sources, "
                "a `actors` key lists the participant actors of the use case, a `steps` key lists the steps about how "
                "the use case works step by step, a `reason` key explaining under what circumstances would the "
                "external system execute this use case.\n"
                '- a "relationship" key lists all the descriptions of relationship among these use cases.\n',
            ],
            stream=False,
        )

        code_blocks = parse_json_code_block(rsp)
        for block in code_blocks:
            detail = ReverseUseCaseDetails.model_validate_json(block)
            await self.graph_db.insert(
                subject=ns_class_name, predicate=GraphKeyword.HAS_CLASS_USE_CASE, object_=detail.model_dump_json()
            )

    @retry(
        wait=wait_random_exponential(min=1, max=20),
        stop=stop_after_attempt(6),
        after=general_after_log(logger),
    )
    async def _rebuild_sequence_view(self, ns_class_name: str):
        """
        Asynchronously reconstructs the sequence diagram for the provided namespace-prefixed class name.

        Args:
            ns_class_name (str): The namespace-prefixed class name for which the sequence diagram is to be reconstructed.
        """
        await self._rebuild_use_case(ns_class_name)

        prompts_blocks = []
        use_case_markdown = await self._get_class_use_cases(ns_class_name)
        if not use_case_markdown:  # external class
            await self.graph_db.insert(subject=ns_class_name, predicate=GraphKeyword.HAS_SEQUENCE_VIEW, object_="")
            return
        block = f"## Use Cases\n{use_case_markdown}"
        prompts_blocks.append(block)

        participants = await self._get_participants(ns_class_name)
        block = "## Participants\n" + "\n".join([f"- {s}" for s in participants])
        prompts_blocks.append(block)

        view = await self._get_uml_class_view(ns_class_name)
        block = "## Mermaid Class Views\n```mermaid\n"
        block += view.get_mermaid()
        block += "\n```\n"
        prompts_blocks.append(block)

        block = "## Source Code\n```python\n"
        block += await self._get_source_code(ns_class_name)
        block += "\n```\n"
        prompts_blocks.append(block)
        prompt = "\n---\n".join(prompts_blocks)

        rsp = await self.llm.aask(
            prompt,
            system_msgs=[
                "You are a Mermaid Sequence Diagram translator in function detail.",
                "Translate the markdown text to a Mermaid Sequence Diagram.",
                "Return a markdown mermaid code block.",
            ],
            stream=False,
        )

        sequence_view = rsp.removeprefix("```mermaid").removesuffix("```")
        await self.graph_db.insert(
            subject=ns_class_name, predicate=GraphKeyword.HAS_SEQUENCE_VIEW, object_=sequence_view
        )

    async def _get_participants(self, ns_class_name: str) -> List[str]:
        """
        Asynchronously returns the participants list of the sequence diagram for the provided namespace-prefixed SPO
        object.

        Args:
            ns_class_name (str): The namespace-prefixed class name for which to retrieve the participants list.

        Returns:
            List[str]: A list of participants in the sequence diagram.
        """
        participants = set()
        detail = await self._get_class_detail(ns_class_name)
        if not detail:
            return []
        participants.update(set(detail.compositions))
        participants.update(set(detail.aggregations))
        return list(participants)

    async def _get_class_use_cases(self, ns_class_name: str) -> str:
        """
        Asynchronously assembles the context about the use case information of the namespace-prefixed SPO object.

        Args:
            ns_class_name (str): The namespace-prefixed class name for which to retrieve use case information.

        Returns:
            str: A string containing the assembled context about the use case information.
        """
        block = ""
        rows = await self.graph_db.select(subject=ns_class_name, predicate=GraphKeyword.HAS_CLASS_USE_CASE)
        for i, r in enumerate(rows):
            detail = ReverseUseCaseDetails.model_validate_json(r.object_)
            block += f"\n### {i + 1}. {detail.description}"
            for j, use_case in enumerate(detail.use_cases):
                block += f"\n#### {i + 1}.{j + 1}. {use_case.description}\n"
                block += "\n##### Inputs\n" + "\n".join([f"- {s}" for s in use_case.inputs])
                block += "\n##### Outputs\n" + "\n".join([f"- {s}" for s in use_case.outputs])
                block += "\n##### Actors\n" + "\n".join([f"- {s}" for s in use_case.actors])
                block += "\n##### Steps\n" + "\n".join([f"- {s}" for s in use_case.steps])
            block += "\n#### Use Case Relationship\n" + "\n".join([f"- {s}" for s in detail.relationship])
        return block + "\n"

    async def _get_class_detail(self, ns_class_name: str) -> DotClassInfo | None:
        """
        Asynchronously retrieves the dot format class details of the namespace-prefixed SPO object.

        Args:
            ns_class_name (str): The namespace-prefixed class name for which to retrieve class details.

        Returns:
            Union[DotClassInfo, None]: A DotClassInfo object representing the dot format class details,
                                       or None if the details are not available.
        """
        rows = await self.graph_db.select(subject=ns_class_name, predicate=GraphKeyword.HAS_DETAIL)
        if not rows:
            return None
        dot_class_info = DotClassInfo.model_validate_json(rows[0].object_)
        return dot_class_info

    async def _get_uml_class_view(self, ns_class_name: str) -> UMLClassView | None:
        """
        Asynchronously retrieves the UML 2.0 format class details of the namespace-prefixed SPO object.

        Args:
            ns_class_name (str): The namespace-prefixed class name for which to retrieve UML class details.

        Returns:
            Union[UMLClassView, None]: A UMLClassView object representing the UML 2.0 format class details,
                                       or None if the details are not available.
        """
        rows = await self.graph_db.select(subject=ns_class_name, predicate=GraphKeyword.HAS_CLASS_VIEW)
        if not rows:
            return None
        class_view = UMLClassView.model_validate_json(rows[0].object_)
        return class_view

    async def _get_source_code(self, ns_class_name: str) -> str:
        """
        Asynchronously retrieves the source code of the namespace-prefixed SPO object.

        Args:
            ns_class_name (str): The namespace-prefixed class name for which to retrieve the source code.

        Returns:
            str: A string containing the source code of the specified namespace-prefixed class.
        """
        rows = await self.graph_db.select(subject=ns_class_name, predicate=GraphKeyword.HAS_PAGE_INFO)
        filename = split_namespace(ns_class_name=ns_class_name)[0]
        if not rows:
            src_filename = RebuildSequenceView._get_full_filename(root=self.i_context, pathname=filename)
            if not src_filename:
                return ""
            return await aread(filename=src_filename, encoding="utf-8")
        code_block_info = CodeBlockInfo.model_validate_json(rows[0].object_)
        return await read_file_block(
            filename=filename, lineno=code_block_info.lineno, end_lineno=code_block_info.end_lineno
        )

    @staticmethod
    def _get_full_filename(root: str | Path, pathname: str | Path) -> Path | None:
        """
        Convert package name to the full path of the module.

        Args:
            root (Union[str, Path]): The root path or string representing the package.
            pathname (Union[str, Path]): The pathname or string representing the module.

        Returns:
            Union[Path, None]: The full path of the module, or None if the path cannot be determined.

        Examples:
            If `root`(workdir) is "/User/xxx/github/MetaGPT/metagpt", and the `pathname` is
            "metagpt/management/skill_manager.py", then the returned value will be
            "/User/xxx/github/MetaGPT/metagpt/management/skill_manager.py"
        """
        if re.match(r"^/.+", pathname):
            return pathname
        files = list_files(root=root)
        postfix = "/" + str(pathname)
        for i in files:
            if str(i).endswith(postfix):
                return i
        return None

    @staticmethod
    def parse_participant(mermaid_sequence_diagram: str) -> List[str]:
        """
        Parses the provided Mermaid sequence diagram and returns the list of participants.

        Args:
            mermaid_sequence_diagram (str): The Mermaid sequence diagram string to be parsed.

        Returns:
            List[str]: A list of participants extracted from the sequence diagram.
        """
        pattern = r"participant ([\w\.]+)"
        matches = re.findall(pattern, mermaid_sequence_diagram)
        matches = [re.sub(r"[\\/'\"]+", "", i) for i in matches]
        return matches

    async def _search_new_participant(self, entry: SPO) -> str | None:
        """
        Asynchronously retrieves a participant whose sequence diagram has not been augmented.

        Args:
            entry (SPO): The SPO object representing the relationship in the graph database.

        Returns:
            Union[str, None]: A participant whose sequence diagram has not been augmented, or None if not found.
        """
        rows = await self.graph_db.select(subject=entry.subject, predicate=GraphKeyword.HAS_SEQUENCE_VIEW)
        if not rows:
            return None
        sequence_view = rows[0].object_
        rows = await self.graph_db.select(subject=entry.subject, predicate=GraphKeyword.HAS_PARTICIPANT)
        merged_participants = []
        for r in rows:
            name = split_namespace(r.object_)[-1]
            merged_participants.append(name)
        participants = self.parse_participant(sequence_view)
        for p in participants:
            if p in merged_participants:
                continue
            return p
        return None

    @retry(
        wait=wait_random_exponential(min=1, max=20),
        stop=stop_after_attempt(6),
        after=general_after_log(logger),
    )
    async def _merge_participant(self, entry: SPO, class_name: str):
        """
        Augments the sequence diagram of `class_name` to the sequence diagram of `entry`.

        Args:
            entry (SPO): The SPO object representing the base sequence diagram.
            class_name (str): The class name whose sequence diagram is to be augmented.
        """
        rows = await self.graph_db.select(predicate=GraphKeyword.IS, object_=GraphKeyword.CLASS)
        participants = []
        for r in rows:
            name = split_namespace(r.subject)[-1]
            if name == class_name:
                participants.append(r)
        if len(participants) == 0:  # external participants
            await self.graph_db.insert(
                subject=entry.subject, predicate=GraphKeyword.HAS_PARTICIPANT, object_=concat_namespace("?", class_name)
            )
            return
        if len(participants) > 1:
            for r in participants:
                await self.graph_db.insert(
                    subject=entry.subject, predicate=GraphKeyword.HAS_PARTICIPANT, object_=auto_namespace(r.subject)
                )
            return

        participant = participants[0]
        await self._rebuild_sequence_view(participant.subject)
        sequence_views = await self.graph_db.select(
            subject=participant.subject, predicate=GraphKeyword.HAS_SEQUENCE_VIEW
        )
        if not sequence_views:  # external class
            return
        rows = await self.graph_db.select(subject=entry.subject, predicate=GraphKeyword.HAS_SEQUENCE_VIEW)
        prompt = f"```mermaid\n{sequence_views[0].object_}\n```\n---\n```mermaid\n{rows[0].object_}\n```"

        rsp = await self.llm.aask(
            prompt,
            system_msgs=[
                "You are a tool to merge sequence diagrams into one.",
                "Participants with the same name are considered identical.",
                "Return the merged Mermaid sequence diagram in a markdown code block format.",
            ],
            stream=False,
        )

        sequence_view = rsp.removeprefix("```mermaid").removesuffix("```")
        rows = await self.graph_db.select(subject=entry.subject, predicate=GraphKeyword.HAS_SEQUENCE_VIEW)
        for r in rows:
            await self.graph_db.delete(subject=r.subject, predicate=r.predicate, object_=r.object_)
        await self.graph_db.insert(
            subject=entry.subject, predicate=GraphKeyword.HAS_SEQUENCE_VIEW, object_=sequence_view
        )
        await self.graph_db.insert(
            subject=entry.subject,
            predicate=GraphKeyword.HAS_SEQUENCE_VIEW_VER,
            object_=concat_namespace(datetime.now().strftime("%Y%m%d%H%M%S%f")[:-3], add_affix(sequence_view)),
        )
        await self.graph_db.insert(
            subject=entry.subject, predicate=GraphKeyword.HAS_PARTICIPANT, object_=auto_namespace(participant.subject)
        )
        await self._save_sequence_view(subject=entry.subject, content=sequence_view)

    async def _save_sequence_view(self, subject: str, content: str):
        pattern = re.compile(r"[^a-zA-Z0-9]")
        name = re.sub(pattern, "_", subject)
        filename = Path(name).with_suffix(".sequence_diagram.mmd")
        await self.context.repo.resources.data_api_design.save(filename=str(filename), content=content)

    async def _search_participants(self, filename: str) -> Set:
        content = await self._get_source_code(filename)

        rsp = await self.llm.aask(
            msg=content,
            system_msgs=[
                "You are a tool for listing all class names used in a source file.",
                "Return a markdown JSON object with: "
                '- a "class_names" key containing the list of class names used in the file; '
                '- a "reasons" key lists all reason objects, each object containing a "class_name" key for class name, a "reference" key explaining the line where the class has been used.',
            ],
        )

        class _Data(BaseModel):
            class_names: List[str]
            reasons: List

        json_blocks = parse_json_code_block(rsp)
        data = _Data.model_validate_json(json_blocks[0])
        return set(data.class_names)



from typing import Any

from pydantic import BaseModel, Field, field_validator

from metagpt.schema import Message
from metagpt.utils.common import any_to_str_set


class RoleExperience(BaseModel):
    id: str = ""
    name: str = ""
    profile: str
    reflection: str
    instruction: str = ""
    response: str
    outcome: str = ""
    round_id: str = ""
    game_setup: str = ""
    version: str = ""

    def rag_key(self) -> str:
        """For search"""
        return self.reflection


class WwMessage(Message):
    # Werewolf Message
    restricted_to: set[str] = Field(default=set(), validate_default=True)

    @field_validator("restricted_to", mode="before")
    @classmethod
    def check_restricted_to(cls, restricted_to: Any):
        return any_to_str_set(restricted_to if restricted_to else set())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Desc   : st' planning execution

import datetime
import math
import random
from typing import Tuple, Union

from metagpt.ext.stanford_town.actions.decide_to_talk import DecideToTalk
from metagpt.ext.stanford_town.actions.gen_action_details import GenActionDetails
from metagpt.ext.stanford_town.actions.gen_daily_schedule import GenDailySchedule
from metagpt.ext.stanford_town.actions.gen_hourly_schedule import GenHourlySchedule
from metagpt.ext.stanford_town.actions.new_decomp_schedule import NewDecompSchedule
from metagpt.ext.stanford_town.actions.summarize_conv import SummarizeConv
from metagpt.ext.stanford_town.actions.task_decomp import TaskDecomp
from metagpt.ext.stanford_town.actions.wake_up import WakeUp
from metagpt.ext.stanford_town.memory.retrieve import new_agent_retrieve
from metagpt.ext.stanford_town.plan.converse import agent_conversation
from metagpt.ext.stanford_town.utils.utils import get_embedding
from metagpt.llm import LLM
from metagpt.logs import logger


async def plan(role: "STRole", roles: dict["STRole"], new_day: bool, retrieved: dict) -> str:
    # PART 1: Generate the hourly schedule.
    if new_day:
        await _long_term_planning(role, new_day)

    # PART 2: If the current action has expired, we want to create a new plan.
    act_check_finished = role.scratch.act_check_finished()
    logger.info(f"Role: {role.name} act_check_finished is {act_check_finished}")
    if act_check_finished:
        await _determine_action(role)

    # PART 3: If you perceived an event that needs to be responded to (saw
    # another role), and retrieved relevant information.
    # Step 1: Retrieved may have multiple events represented in it. The first
    #         job here is to determine which of the events we want to focus
    #         on for the role.
    #         <focused_event> takes the form of a dictionary like this:
    #         dictionary {["curr_event"] = <ConceptNode>,
    #                     ["events"] = [<ConceptNode>, ...],
    #                     ["thoughts"] = [<ConceptNode>, ...]}
    focused_event = False
    if retrieved.keys():
        focused_event = _choose_retrieved(role.name, retrieved)

    # Step 2: Once we choose an event, we need to determine whether the
    #         role will take any actions for the perceived event. There are
    #         three possible modes of reaction returned by _should_react.
    #         a) "chat with {target_role.name}"
    #         b) "react"
    #         c) False
    logger.info(f"Role: {role.name} focused_event: {focused_event}")
    if focused_event:
        reaction_mode = await _should_react(role, focused_event, roles)
        logger.info(f"Role: {role.name} reaction_mode: {reaction_mode}")
        if reaction_mode:
            # If we do want to chat, then we generate conversation
            if reaction_mode[:9] == "chat with":
                await _chat_react(role, reaction_mode, roles)
            elif reaction_mode[:4] == "wait":
                await _wait_react(role, reaction_mode)

    # Step 3: Chat-related state clean up.
    # If the persona is not chatting with anyone, we clean up any of the
    # chat-related states here.
    if role.rc.scratch.act_event[1] != "chat with":
        role.rc.scratch.chatting_with = None
        role.rc.scratch.chat = None
        role.rc.scratch.chatting_end_time = None
    # We want to make sure that the persona does not keep conversing with each
    # other in an infinite loop. So, chatting_with_buffer maintains a form of
    # buffer that makes the persona wait from talking to the same target
    # immediately after chatting once. We keep track of the buffer value here.
    curr_persona_chat_buffer = role.rc.scratch.chatting_with_buffer
    for persona_name, buffer_count in curr_persona_chat_buffer.items():
        if persona_name != role.rc.scratch.chatting_with:
            role.rc.scratch.chatting_with_buffer[persona_name] -= 1

    return role.rc.scratch.act_address


def _choose_retrieved(role_name: str, retrieved: dict) -> Union[None, dict]:
    """
    Retrieved elements have multiple core "curr_events". We need to choose one
    event to which we are going to react to. We pick that event here.
    Args:
      role_name: Current role instance's name whose action we are determining.
      retrieved: A dictionary of <ConceptNode> that were retrieved from the
                 the role's associative memory. This dictionary takes the
                 following form:
                 dictionary[event.description] =
                   {["curr_event"] = <ConceptNode>,
                    ["events"] = [<ConceptNode>, ...],
                    ["thoughts"] = [<ConceptNode>, ...] }
    """
    # Once we are done with the reflection, we might want to build a more
    # complex structure here.

    # We do not want to take self events... for now
    copy_retrieved = retrieved.copy()
    for event_desc, rel_ctx in copy_retrieved.items():
        curr_event = rel_ctx["curr_event"]
        if curr_event.subject == role_name:
            del retrieved[event_desc]

    # Always choose role first.
    priority = []
    for event_desc, rel_ctx in retrieved.items():
        curr_event = rel_ctx["curr_event"]
        if ":" not in curr_event.subject and curr_event.subject != role_name:
            priority += [rel_ctx]
    if priority:
        return random.choice(priority)

    # Skip idle.
    for event_desc, rel_ctx in retrieved.items():
        if "is idle" not in event_desc:
            priority += [rel_ctx]
    if priority:
        return random.choice(priority)
    return None


async def _should_react(role: "STRole", retrieved: dict, roles: dict):
    """
    Determines what form of reaction the role should exihibit given the
    retrieved values.
    INPUT
      role: Current <"STRole"> instance whose action we are determining.
      retrieved: A dictionary of <ConceptNode> that were retrieved from the
                 the role's associative memory. This dictionary takes the
                 following form:
                 dictionary[event.description] =
                   {["curr_event"] = <ConceptNode>,
                    ["events"] = [<ConceptNode>, ...],
                    ["thoughts"] = [<ConceptNode>, ...] }
      roles: A dictionary that contains all role names as keys, and the
                <"STRole"> instance as values.
    """

    async def lets_talk(init_role: "STRole", target_role: "STRole", retrieved: dict):
        if init_role.name == target_role.name:
            logger.info(f"Role: {role.name} _should_react lets_talk meet same role, return False")
            return False

        scratch = init_role.rc.scratch
        target_scratch = target_role.rc.scratch
        if (
            not target_scratch.act_address
            or not target_scratch.act_description
            or not scratch.act_address
            or not scratch.act_description
        ):
            return False

        if "sleeping" in target_scratch.act_description or "sleeping" in scratch.act_description:
            return False

        if scratch.curr_time.hour == 23:
            return False

        if "<waiting>" in target_scratch.act_address:
            return False

        if target_scratch.chatting_with or scratch.chatting_with:
            return False

        if target_role.name in scratch.chatting_with_buffer:
            if scratch.chatting_with_buffer[target_role.name] > 0:
                return False

        if await DecideToTalk().run(init_role, target_role, retrieved):
            return True

        return False

    async def lets_react(init_role: "STRole", target_role: "STRole", retrieved: dict):
        if init_role.name == target_role.name:
            logger.info(f"Role: {role.name} _should_react lets_react meet same role, return False")
            return False

        scratch = init_role.rc.scratch
        target_scratch = target_role.rc.scratch
        if (
            not target_scratch.act_address
            or not target_scratch.act_description
            or not scratch.act_address
            or not scratch.act_description
        ):
            return False

        if "sleeping" in target_scratch.act_description or "sleeping" in scratch.act_description:
            return False

        # return False
        if scratch.curr_time.hour == 23:
            return False

        if "waiting" in target_scratch.act_description:
            return False
        if scratch.planned_path == []:
            return False

        if scratch.act_address != target_scratch.act_address:
            return False

        react_mode = await DecideToTalk().run(init_role, target_role, retrieved)

        if react_mode == "1":
            wait_until = (
                target_scratch.act_start_time + datetime.timedelta(minutes=target_scratch.act_duration - 1)
            ).strftime("%B %d, %Y, %H:%M:%S")
            return f"wait: {wait_until}"
        elif react_mode == "2":
            return False
            return "do other things"
        else:
            return False  # "keep"

    # If the role is chatting right now, default to no reaction
    scratch = role.rc.scratch
    if scratch.chatting_with:
        return False
    if "<waiting>" in scratch.act_address:
        return False

    # Recall that retrieved takes the following form:
    # dictionary {["curr_event"] = <ConceptNode>}
    curr_event = retrieved["curr_event"]
    logger.info(f"Role: {role.name} _should_react curr_event.subject: {curr_event.subject}")

    if ":" not in curr_event.subject:
        # this is a role event.
        if await lets_talk(role, roles[curr_event.subject], retrieved):
            return f"chat with {curr_event.subject}"
        react_mode = await lets_react(role, roles[curr_event.subject], retrieved)
        return react_mode
    return False


async def _chat_react(role: "STRole", reaction_mode: str, roles: dict["STRole"]):
    # There are two roles -- the role who is initiating the conversation
    # and the role who is the target. We get the role instances here.
    init_role = role
    target_role = roles[reaction_mode[9:].strip()]

    # Actually creating the conversation here.
    convo, duration_min = await generate_convo(init_role, target_role)  # 2222
    convo_summary = await generate_convo_summary(convo)
    inserted_act = convo_summary
    inserted_act_dur = duration_min

    act_start_time = target_role.rc.scratch.act_start_time

    curr_time = target_role.rc.scratch.curr_time
    if curr_time.second != 0:
        temp_curr_time = curr_time + datetime.timedelta(seconds=60 - curr_time.second)
        chatting_end_time = temp_curr_time + datetime.timedelta(minutes=inserted_act_dur)
    else:
        chatting_end_time = curr_time + datetime.timedelta(minutes=inserted_act_dur)

    for role, p in [("init", init_role), ("target", target_role)]:
        if role == "init":
            act_address = f"<persona> {target_role.name}"
            act_event = (p.name, "chat with", target_role.name)
            chatting_with = target_role.name
            chatting_with_buffer = {}
            chatting_with_buffer[target_role.name] = 800
        elif role == "target":
            act_address = f"<persona> {init_role.name}"
            act_event = (p.name, "chat with", init_role.name)
            chatting_with = init_role.name
            chatting_with_buffer = {}
            chatting_with_buffer[init_role.name] = 800

        act_pronunciatio = "💬"
        act_obj_description = None
        act_obj_pronunciatio = None
        act_obj_event = (None, None, None)

        await _create_react(
            p,
            inserted_act,
            inserted_act_dur,
            act_address,
            act_event,
            chatting_with,
            convo,
            chatting_with_buffer,
            chatting_end_time,
            act_pronunciatio,
            act_obj_description,
            act_obj_pronunciatio,
            act_obj_event,
            act_start_time,
        )


async def _create_react(
    role: "STRole",
    inserted_act: str,
    inserted_act_dur: int,
    act_address: str,
    act_event: Tuple,
    chatting_with: str,
    chat: list,
    chatting_with_buffer: dict,
    chatting_end_time: datetime,
    act_pronunciatio: str,
    act_obj_description: str,
    act_obj_pronunciatio: str,
    act_obj_event: Tuple,
    act_start_time=None,
):
    p = role
    scratch = role.rc.scratch

    min_sum = 0
    for i in range(scratch.get_f_daily_schedule_hourly_org_index()):
        min_sum += scratch.f_daily_schedule_hourly_org[i][1]
    start_hour = int(min_sum / 60)

    if scratch.f_daily_schedule_hourly_org[scratch.get_f_daily_schedule_hourly_org_index()][1] >= 120:
        end_hour = (
            start_hour + scratch.f_daily_schedule_hourly_org[scratch.get_f_daily_schedule_hourly_org_index()][1] / 60
        )

    elif (
        scratch.f_daily_schedule_hourly_org[scratch.get_f_daily_schedule_hourly_org_index()][1]
        + scratch.f_daily_schedule_hourly_org[scratch.get_f_daily_schedule_hourly_org_index() + 1][1]
    ):
        end_hour = start_hour + (
            (
                scratch.f_daily_schedule_hourly_org[scratch.get_f_daily_schedule_hourly_org_index()][1]
                + scratch.f_daily_schedule_hourly_org[scratch.get_f_daily_schedule_hourly_org_index() + 1][1]
            )
            / 60
        )

    else:
        end_hour = start_hour + 2
    end_hour = int(end_hour)

    dur_sum = 0
    count = 0
    start_index = None
    end_index = None
    for act, dur in scratch.f_daily_schedule:
        if dur_sum >= start_hour * 60 and start_index is None:
            start_index = count
        if dur_sum >= end_hour * 60 and end_index is None:
            end_index = count
        dur_sum += dur
        count += 1

    ret = await generate_new_decomp_schedule(p, inserted_act, inserted_act_dur, start_hour, end_hour)
    scratch.f_daily_schedule[start_index:end_index] = ret
    scratch.add_new_action(
        act_address,
        inserted_act_dur,
        inserted_act,
        act_pronunciatio,
        act_event,
        chatting_with,
        chat,
        chatting_with_buffer,
        chatting_end_time,
        act_obj_description,
        act_obj_pronunciatio,
        act_obj_event,
        act_start_time,
    )


async def _wait_react(role: "STRole", reaction_mode: str):
    scratch = role.rc.scratch

    inserted_act = f'waiting to start {scratch.act_description.split("(")[-1][:-1]}'
    end_time = datetime.datetime.strptime(reaction_mode[6:].strip(), "%B %d, %Y, %H:%M:%S")
    inserted_act_dur = (
        (end_time.minute + end_time.hour * 60) - (scratch.curr_time.minute + scratch.curr_time.hour * 60) + 1
    )

    act_address = f"<waiting> {scratch.curr_tile[0]} {scratch.curr_tile[1]}"
    act_event = (role.name, "waiting to start", scratch.act_description.split("(")[-1][:-1])
    chatting_with = None
    chat = None
    chatting_with_buffer = None
    chatting_end_time = None

    act_pronunciatio = "⌛"
    act_obj_description = None
    act_obj_pronunciatio = None
    act_obj_event = (None, None, None)

    await _create_react(
        role,
        inserted_act,
        inserted_act_dur,
        act_address,
        act_event,
        chatting_with,
        chat,
        chatting_with_buffer,
        chatting_end_time,
        act_pronunciatio,
        act_obj_description,
        act_obj_pronunciatio,
        act_obj_event,
    )


async def generate_convo(init_role: "STRole", target_role: "STRole") -> Union[list, int]:
    convo = await agent_conversation(init_role, target_role)
    all_utt = ""

    for row in convo:
        speaker = row[0]
        utt = row[1]
        all_utt += f"{speaker}: {utt}\n"

    convo_length = math.ceil(int(len(all_utt) / 8) / 30)

    return convo, convo_length


async def generate_convo_summary(conv: list[list[str]]) -> str:
    conv_summary = await SummarizeConv().run(conv)
    return conv_summary


async def generate_new_decomp_schedule(
    role: "STRole", inserted_act: str, inserted_act_dur: int, start_hour: int, end_hour: int
):
    # Step 1: Setting up the core variables for the function.
    # <p> is the role whose schedule we are editing right now.
    scratch = role.rc.scratch
    # <today_min_pass> indicates the number of minutes that have passed today.
    today_min_pass = int(scratch.curr_time.hour) * 60 + int(scratch.curr_time.minute) + 1

    # Step 2: We need to create <main_act_dur> and <truncated_act_dur>.
    main_act_dur = []
    truncated_act_dur = []
    dur_sum = 0  # duration sum
    count = 0  # enumerate count
    truncated_fin = False

    logger.debug(f"DEBUG::: {scratch.name}")
    for act, dur in scratch.f_daily_schedule:
        if (dur_sum >= start_hour * 60) and (dur_sum < end_hour * 60):
            main_act_dur += [[act, dur]]
            if dur_sum <= today_min_pass:
                truncated_act_dur += [[act, dur]]
            elif dur_sum > today_min_pass and not truncated_fin:
                # We need to insert that last act, duration list like this one:
                # e.g., ['wakes up and completes her morning routine (wakes up...)', 2]
                truncated_act_dur += [[scratch.f_daily_schedule[count][0], dur_sum - today_min_pass]]
                truncated_act_dur[-1][-1] -= (
                    dur_sum - today_min_pass
                )  # DEC 7 DEBUG;.. is the +1 the right thing to do???
                # DEC 7 DEBUG;.. is the +1 the right thing to do???
                # truncated_act_dur[-1][-1] -= (dur_sum - today_min_pass + 1)
                logger.debug(f"DEBUG::: {truncated_act_dur}")

                # DEC 7 DEBUG;.. is the +1 the right thing to do???
                # truncated_act_dur[-1][-1] -= (dur_sum - today_min_pass)
                truncated_fin = True
        dur_sum += dur
        count += 1

    main_act_dur = main_act_dur

    x = (
        truncated_act_dur[-1][0].split("(")[0].strip()
        + " (on the way to "
        + truncated_act_dur[-1][0].split("(")[-1][:-1]
        + ")"
    )
    truncated_act_dur[-1][0] = x

    if "(" in truncated_act_dur[-1][0]:
        inserted_act = truncated_act_dur[-1][0].split("(")[0].strip() + " (" + inserted_act + ")"

    # To do inserted_act_dur+1 below is an important decision but I'm not sure
    # if I understand the full extent of its implications. Might want to
    # revisit.
    truncated_act_dur += [[inserted_act, inserted_act_dur]]
    start_time_hour = datetime.datetime(2022, 10, 31, 0, 0) + datetime.timedelta(hours=start_hour)
    end_time_hour = datetime.datetime(2022, 10, 31, 0, 0) + datetime.timedelta(hours=end_hour)

    return await NewDecompSchedule().run(
        role, main_act_dur, truncated_act_dur, start_time_hour, end_time_hour, inserted_act, inserted_act_dur
    )


async def _long_term_planning(role: "STRole", new_day: bool):
    """
    Formulates the role's daily long-term plan if it is the start of a new
    day. This basically has two components: first, we create the wake-up hour,
    and second, we create the hourly schedule based on it.
    INPUT
        new_day: Indicates whether the current time signals a "First day",
                "New day", or False (for neither). This is important because we
                create the roles' long term planning on the new day.
    """
    # We start by creating the wake up hour for the role.
    wake_up_hour = await WakeUp().run(role)
    wake_up_hour = int(wake_up_hour)
    logger.info(f"Role: {role.name} long_term_planning, wake_up_hour: {wake_up_hour}")

    # When it is a new day, we start by creating the daily_req of the role.
    # Note that the daily_req is a list of strings that describe the role's
    # day in broad strokes.
    if new_day == "First day":
        # Bootstrapping the daily plan for the start of then generation:
        # if this is the start of generation (so there is no previous day's
        # daily requirement, or if we are on a new day, we want to create a new
        # set of daily requirements.
        role.scratch.daily_req = await GenDailySchedule().run(role, wake_up_hour)
        logger.info(f"Role: {role.name} daily requirements: {role.scratch.daily_req}")
    elif new_day == "New day":
        revise_identity(role)

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - TODO
        # We need to create a new daily_req here...
        role.scratch.daily_req = role.scratch.daily_req

    # Based on the daily_req, we create an hourly schedule for the role,
    # which is a list of todo items with a time duration (in minutes) that
    # add up to 24 hours.
    role.scratch.f_daily_schedule = await GenHourlySchedule().run(role, wake_up_hour)
    logger.info(f"Role: {role.name} f_daily_schedule: {role.scratch.f_daily_schedule}")
    role.scratch.f_daily_schedule_hourly_org = role.scratch.f_daily_schedule[:]

    # Added March 4 -- adding plan to the memory.
    thought = f"This is {role.scratch.name}'s plan for {role.scratch.curr_time.strftime('%A %B %d')}:"
    for i in role.scratch.daily_req:
        thought += f" {i},"
    thought = thought[:-1] + "."
    created = role.scratch.curr_time
    expiration = role.scratch.curr_time + datetime.timedelta(days=30)
    s, p, o = (role.scratch.name, "plan", role.scratch.curr_time.strftime("%A %B %d"))
    keywords = set(["plan"])
    thought_poignancy = 5
    thought_embedding_pair = (thought, get_embedding(thought))
    role.a_mem.add_thought(
        created, expiration, s, p, o, thought, keywords, thought_poignancy, thought_embedding_pair, None
    )


async def _determine_action(role: "STRole"):
    """
    Creates the next action sequence for the role.
    The main goal of this function is to run "add_new_action" on the role's
    scratch space, which sets up all the action related variables for the next
    action.
    As a part of this, the role may need to decompose its hourly schedule as
    needed.
    INPUT
        role: Current <Persona> instance whose action we are determining.
    """

    def determine_decomp(act_desp, act_dura):
        """
        Given an action description and its duration, we determine whether we need
        to decompose it. If the action is about the agent sleeping, we generally
        do not want to decompose it, so that's what we catch here.

        INPUT:
        act_desp: the description of the action (e.g., "sleeping")
        act_dura: the duration of the action in minutes.
        OUTPUT:
        a boolean. True if we need to decompose, False otherwise.
        """
        if "sleep" not in act_desp and "bed" not in act_desp:
            return True
        elif "sleeping" in act_desp or "asleep" in act_desp or "in bed" in act_desp:
            return False
        elif "sleep" in act_desp or "bed" in act_desp:
            if act_dura > 60:
                return False
        return True

    # The goal of this function is to get us the action associated with
    # <curr_index>. As a part of this, we may need to decompose some large
    # chunk actions.
    # Importantly, we try to decompose at least two hours worth of schedule at
    # any given point.
    curr_index = role.scratch.get_f_daily_schedule_index()
    curr_index_60 = role.scratch.get_f_daily_schedule_index(advance=60)

    logger.info(f"f_daily_schedule: {role.scratch.f_daily_schedule}")
    # * Decompose *
    # During the first hour of the day, we need to decompose two hours
    # sequence. We do that here.
    if curr_index == 0:
        # This portion is invoked if it is the first hour of the day.
        act_desp, act_dura = role.scratch.f_daily_schedule[curr_index]
        if act_dura >= 60:
            # We decompose if the next action is longer than an hour, and fits the
            # criteria described in determine_decomp.
            if determine_decomp(act_desp, act_dura):
                role.scratch.f_daily_schedule[curr_index : curr_index + 1] = await TaskDecomp().run(
                    role, act_desp, act_dura
                )
        if curr_index_60 + 1 < len(role.scratch.f_daily_schedule):
            act_desp, act_dura = role.scratch.f_daily_schedule[curr_index_60 + 1]
            if act_dura >= 60:
                if determine_decomp(act_desp, act_dura):
                    role.scratch.f_daily_schedule[curr_index_60 + 1 : curr_index_60 + 2] = await TaskDecomp().run(
                        role, act_desp, act_dura
                    )

    if curr_index_60 < len(role.scratch.f_daily_schedule):
        # If it is not the first hour of the day, this is always invoked (it is
        # also invoked during the first hour of the day -- to double up so we can
        # decompose two hours in one go). Of course, we need to have something to
        # decompose as well, so we check for that too.
        if role.scratch.curr_time.hour < 23:
            # And we don't want to decompose after 11 pm.
            act_desp, act_dura = role.scratch.f_daily_schedule[curr_index_60]
            if act_dura >= 60:
                if determine_decomp(act_desp, act_dura):
                    role.scratch.f_daily_schedule[curr_index_60 : curr_index_60 + 1] = await TaskDecomp().run(
                        role, act_desp, act_dura
                    )
    # * End of Decompose *

    # Generate an <Action> instance from the action description and duration. By
    # this point, we assume that all the relevant actions are decomposed and
    # ready in f_daily_schedule.
    logger.debug("DEBUG LJSDLFSKJF")
    for i in role.scratch.f_daily_schedule:
        logger.debug(i)
    logger.debug(curr_index)
    logger.debug(len(role.scratch.f_daily_schedule))
    logger.debug(role.scratch.name)

    # 1440
    x_emergency = 0
    for i in role.scratch.f_daily_schedule:
        x_emergency += i[1]

    if 1440 - x_emergency > 0:
        logger.info(f"x_emergency__AAA: {x_emergency}")
    role.scratch.f_daily_schedule += [["sleeping", 1440 - x_emergency]]

    act_desp, act_dura = role.scratch.f_daily_schedule[curr_index]

    new_action_details = await GenActionDetails().run(role, act_desp, act_dura)
    # Adding the action to role's queue.
    role.scratch.add_new_action(**new_action_details)


def revise_identity(role: "STRole"):
    p_name = role.scratch.name

    focal_points = [
        f"{p_name}'s plan for {role.scratch.get_str_curr_date_str()}.",
        f"Important recent events for {p_name}'s life.",
    ]
    retrieved = new_agent_retrieve(role, focal_points)

    statements = "[Statements]\n"
    for key, val in retrieved.items():
        for i in val:
            statements += f"{i.created.strftime('%A %B %d -- %H:%M %p')}: {i.embedding_key}\n"

    plan_prompt = statements + "\n"
    plan_prompt += f"Given the statements above, is there anything that {p_name} should remember as they plan for"
    plan_prompt += f" *{role.scratch.curr_time.strftime('%A %B %d')}*? "
    plan_prompt += "If there is any scheduling information, be as specific as possible (include date, time, and location if stated in the statement)\n\n"
    plan_prompt += f"Write the response from {p_name}'s perspective."
    plan_note = LLM().ask(plan_prompt)

    thought_prompt = statements + "\n"
    thought_prompt += (
        f"Given the statements above, how might we summarize {p_name}'s feelings about their days up to now?\n\n"
    )
    thought_prompt += f"Write the response from {p_name}'s perspective."
    thought_note = LLM().ask(thought_prompt)

    currently_prompt = (
        f"{p_name}'s status from {(role.scratch.curr_time - datetime.timedelta(days=1)).strftime('%A %B %d')}:\n"
    )
    currently_prompt += f"{role.scratch.currently}\n\n"
    currently_prompt += f"{p_name}'s thoughts at the end of {(role.scratch.curr_time - datetime.timedelta(days=1)).strftime('%A %B %d')}:\n"
    currently_prompt += (plan_note + thought_note).replace("\n", "") + "\n\n"
    currently_prompt += f"It is now {role.scratch.curr_time.strftime('%A %B %d')}. Given the above, write {p_name}'s status for {role.scratch.curr_time.strftime('%A %B %d')} that reflects {p_name}'s thoughts at the end of {(role.scratch.curr_time - datetime.timedelta(days=1)).strftime('%A %B %d')}. Write this in third-person talking about {p_name}."
    currently_prompt += "If there is any scheduling information, be as specific as possible (include date, time, and location if stated in the statement).\n\n"
    currently_prompt += "Follow this format below:\nStatus: <new status>"
    new_currently = LLM().ask(currently_prompt)

    role.scratch.currently = new_currently

    daily_req_prompt = role.scratch.get_str_iss() + "\n"
    daily_req_prompt += f"Today is {role.scratch.curr_time.strftime('%A %B %d')}. Here is {role.scratch.name}'s plan today in broad-strokes (with the time of the day. e.g., have a lunch at 12:00 pm, watch TV from 7 to 8 pm).\n\n"
    daily_req_prompt += "Follow this format (the list should have 4~6 items but no more):\n"
    daily_req_prompt += "1. wake up and complete the morning routine at <time>, 2. ..."

    new_daily_req = LLM().ask(daily_req_prompt)
    new_daily_req = new_daily_req.replace("\n", " ")
    role.scratch.daily_plan_req = new_daily_req



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Desc   : BasicMemory,AgentMemory实现

from datetime import datetime
from pathlib import Path
from typing import Optional

from pydantic import Field, field_serializer, model_validator

from metagpt.logs import logger
from metagpt.memory.memory import Memory
from metagpt.schema import Message
from metagpt.utils.common import read_json_file, write_json_file


class BasicMemory(Message):
    """
    BasicMemory继承于MG的Message类，其中content属性替代description属性
    Message类中对于Chat类型支持的非常好，对于Agent个体的Perceive,Reflection,Plan支持的并不多
    在Type设计上，我们延续GA的三个种类，但是对于Chat种类的对话进行特别设计（具体怎么设计还没想好）
    """

    memory_id: Optional[str] = Field(default=None)  # 记忆ID
    memory_count: int = -1  # 第几个记忆，实际数值与Memory相等
    type_count: int = -1  # 第几种记忆，类型为整数
    memory_type: Optional[str] = Field(default=None)  # 记忆类型，包含 event,thought,chat三种类型
    depth: int = -1  # 记忆深度，类型为整数
    created: Optional[datetime] = Field(default=None)  # 创建时间
    expiration: Optional[datetime] = Field(default=None)  # 记忆失效时间，默认为空（）
    last_accessed: Optional[datetime] = Field(default=None)  # 上一次调用的时间，初始化时候与self.created一致
    subject: Optional[str] = Field(default=None)  # 主语
    predicate: Optional[str] = Field(default=None)  # 谓语
    object: Optional[str] = Field(default=None)  # 宾语

    description: Optional[str] = Field(default=None)
    embedding_key: Optional[str] = Field(default=None)  # 内容与self.content一致
    poignancy: int = -1  # importance值
    keywords: list[str] = Field(default=[])  # keywords
    filling: list = Field(default=[])  # 装的与之相关联的memory_id的列表

    __hash__ = object.__hash__  # support hash in AgentMemory

    @model_validator(mode="before")
    @classmethod
    def check_values(cls, values):
        if "created" in values:
            values["last_accessed"] = values["created"]
        if "content" in values:
            values["description"] = values["content"]
        if "filling" in values:
            values["filling"] = values["filling"] or []
        return values

    @field_serializer("created", "expiration")
    def transform_time_field(self, time_field: Optional[datetime]) -> str:
        if time_field:
            time_field = time_field.strftime("%Y-%m-%d %H:%M:%S")
        return time_field

    def summary(self):
        return self.subject, self.predicate, self.object

    def save_to_dict(self) -> dict:
        """
        将MemoryBasic类转化为字典，用于存储json文件
        这里需要注意，cause_by跟GA不兼容，所以需要做一个格式转换
        """
        memory_dict = dict()
        node_id = self.memory_id
        basic_mem_obj = self.model_dump(
            include=[
                "node_count",
                "type_count",
                "type",
                "depth",
                "created",
                "expiration",
                "subject",
                "predicate",
                "object",
                "description",
                "embedding_key",
                "poignancy",
                "keywords",
                "filling",
                "cause_by",
            ]
        )

        memory_dict[node_id] = basic_mem_obj
        return memory_dict


class AgentMemory(Memory):
    """
    GA中主要存储三种JSON
    1. embedding.json (Dict embedding_key:embedding)
    2. Node.json (Dict Node_id:Node)
    3. kw_strength.json
    """

    storage: list[BasicMemory] = []  # 重写Storage，存储BasicMemory所有节点
    event_list: list[BasicMemory] = []  # 存储event记忆
    thought_list: list[BasicMemory] = []  # 存储thought记忆
    chat_list: list[BasicMemory] = []  # chat-related memory

    event_keywords: dict[str, list[BasicMemory]] = dict()  # 存储keywords
    thought_keywords: dict[str, list[BasicMemory]] = dict()
    chat_keywords: dict[str, list[BasicMemory]] = dict()

    kw_strength_event: dict[str, int] = dict()
    kw_strength_thought: dict[str, int] = dict()

    memory_saved: Optional[Path] = Field(default=None)
    embeddings: dict[str, list[float]] = dict()

    def set_mem_path(self, memory_saved: Path):
        self.memory_saved = memory_saved
        self.load(memory_saved)

    def save(self, memory_saved: Path):
        """
        将MemoryBasic类存储为Nodes.json形式。复现GA中的Kw Strength.json形式
        这里添加一个路径即可
        TODO 这里在存储时候进行倒序存储，之后需要验证（test_memory通过）
        """
        memory_json = dict()
        for i in range(len(self.storage)):
            memory_node = self.storage[len(self.storage) - i - 1]
            memory_node = memory_node.save_to_dict()
            memory_json.update(memory_node)
        write_json_file(memory_saved.joinpath("nodes.json"), memory_json)
        write_json_file(memory_saved.joinpath("embeddings.json"), self.embeddings)

        strength_json = dict()
        strength_json["kw_strength_event"] = self.kw_strength_event
        strength_json["kw_strength_thought"] = self.kw_strength_thought
        write_json_file(memory_saved.joinpath("kw_strength.json"), strength_json)

    def load(self, memory_saved: Path):
        """
        将GA的JSON解析，填充到AgentMemory类之中
        """
        self.embeddings = read_json_file(memory_saved.joinpath("embeddings.json"))
        memory_load = read_json_file(memory_saved.joinpath("nodes.json"))
        for count in range(len(memory_load.keys())):
            node_id = f"node_{str(count + 1)}"
            node_details = memory_load[node_id]
            node_type = node_details["type"]
            created = datetime.strptime(node_details["created"], "%Y-%m-%d %H:%M:%S")
            expiration = None
            if node_details["expiration"]:
                expiration = datetime.strptime(node_details["expiration"], "%Y-%m-%d %H:%M:%S")

            s = node_details["subject"]
            p = node_details["predicate"]
            o = node_details["object"]

            description = node_details["description"]
            embedding_pair = (node_details["embedding_key"], self.embeddings[node_details["embedding_key"]])
            poignancy = node_details["poignancy"]
            keywords = set(node_details["keywords"])
            filling = node_details["filling"]
            if node_type == "thought":
                self.add_thought(
                    created, expiration, s, p, o, description, keywords, poignancy, embedding_pair, filling
                )
            if node_type == "event":
                self.add_event(created, expiration, s, p, o, description, keywords, poignancy, embedding_pair, filling)
            if node_type == "chat":
                self.add_chat(created, expiration, s, p, o, description, keywords, poignancy, embedding_pair, filling)

        strength_keywords_load = read_json_file(memory_saved.joinpath("kw_strength.json"))
        if strength_keywords_load["kw_strength_event"]:
            self.kw_strength_event = strength_keywords_load["kw_strength_event"]
        if strength_keywords_load["kw_strength_thought"]:
            self.kw_strength_thought = strength_keywords_load["kw_strength_thought"]

    def add(self, memory_basic: BasicMemory):
        """
        Add a new message to storage, while updating the index
        重写add方法，修改原有的Message类为BasicMemory类，并添加不同的记忆类型添加方式
        """
        if memory_basic.memory_id in self.storage:
            return
        self.storage.append(memory_basic)
        if memory_basic.memory_type == "chat":
            self.chat_list[0:0] = [memory_basic]
            return
        if memory_basic.memory_type == "thought":
            self.thought_list[0:0] = [memory_basic]
            return
        if memory_basic.memory_type == "event":
            self.event_list[0:0] = [memory_basic]
            return

    def add_chat(
        self, created, expiration, s, p, o, content, keywords, poignancy, embedding_pair, filling, cause_by=""
    ):
        """
        调用add方法，初始化chat，在创建的时候就需要调用embedding函数
        """
        memory_count = len(self.storage) + 1
        type_count = len(self.thought_list) + 1
        memory_type = "chat"
        memory_id = f"node_{str(memory_count)}"
        depth = 1

        memory_node = BasicMemory(
            memory_id=memory_id,
            memory_count=memory_count,
            type_count=type_count,
            memory_type=memory_type,
            depth=depth,
            created=created,
            expiration=expiration,
            subject=s,
            predicate=p,
            object=o,
            description=content,
            embedding_key=embedding_pair[0],
            poignancy=poignancy,
            keywords=keywords,
            filling=filling,
            cause_by=cause_by,
        )

        keywords = [i.lower() for i in keywords]
        for kw in keywords:
            if kw in self.chat_keywords:
                self.chat_keywords[kw][0:0] = [memory_node]
            else:
                self.chat_keywords[kw] = [memory_node]

        self.add(memory_node)

        self.embeddings[embedding_pair[0]] = embedding_pair[1]
        return memory_node

    def add_thought(self, created, expiration, s, p, o, content, keywords, poignancy, embedding_pair, filling):
        """
        调用add方法，初始化thought
        """
        memory_count = len(self.storage) + 1
        type_count = len(self.thought_list) + 1
        memory_type = "thought"
        memory_id = f"node_{str(memory_count)}"
        depth = 1

        try:
            if filling:
                depth_list = [memory_node.depth for memory_node in self.storage if memory_node.memory_id in filling]
                depth += max(depth_list)
        except Exception as exp:
            logger.warning(f"filling init occur {exp}")
            pass

        memory_node = BasicMemory(
            memory_id=memory_id,
            memory_count=memory_count,
            type_count=type_count,
            memory_type=memory_type,
            depth=depth,
            created=created,
            expiration=expiration,
            subject=s,
            predicate=p,
            object=o,
            description=content,
            embedding_key=embedding_pair[0],
            poignancy=poignancy,
            keywords=keywords,
            filling=filling,
        )

        keywords = [i.lower() for i in keywords]
        for kw in keywords:
            if kw in self.thought_keywords:
                self.thought_keywords[kw][0:0] = [memory_node]
            else:
                self.thought_keywords[kw] = [memory_node]

        self.add(memory_node)

        if f"{p} {o}" != "is idle":
            for kw in keywords:
                if kw in self.kw_strength_thought:
                    self.kw_strength_thought[kw] += 1
                else:
                    self.kw_strength_thought[kw] = 1

        self.embeddings[embedding_pair[0]] = embedding_pair[1]
        return memory_node

    def add_event(self, created, expiration, s, p, o, content, keywords, poignancy, embedding_pair, filling):
        """
        调用add方法，初始化event
        """
        memory_count = len(self.storage) + 1
        type_count = len(self.event_list) + 1
        memory_type = "event"
        memory_id = f"node_{str(memory_count)}"
        depth = 0

        if "(" in content:
            content = " ".join(content.split()[:3]) + " " + content.split("(")[-1][:-1]

        memory_node = BasicMemory(
            memory_id=memory_id,
            memory_count=memory_count,
            type_count=type_count,
            memory_type=memory_type,
            depth=depth,
            created=created,
            expiration=expiration,
            subject=s,
            predicate=p,
            object=o,
            description=content,
            embedding_key=embedding_pair[0],
            poignancy=poignancy,
            keywords=keywords,
            filling=filling,
        )

        keywords = [i.lower() for i in keywords]
        for kw in keywords:
            if kw in self.event_keywords:
                self.event_keywords[kw][0:0] = [memory_node]
            else:
                self.event_keywords[kw] = [memory_node]

        self.add(memory_node)

        if f"{p} {o}" != "is idle":
            for kw in keywords:
                if kw in self.kw_strength_event:
                    self.kw_strength_event[kw] += 1
                else:
                    self.kw_strength_event[kw] = 1

        self.embeddings[embedding_pair[0]] = embedding_pair[1]
        return memory_node

    def get_summarized_latest_events(self, retention):
        ret_set = set()
        for e_node in self.event_list[:retention]:
            ret_set.add(e_node.summary())
        return ret_set

    def get_last_chat(self, target_role_name: str):
        if target_role_name.lower() in self.chat_keywords:
            return self.chat_keywords[target_role_name.lower()][0]
        else:
            return False

    def retrieve_relevant_thoughts(self, s_content: str, p_content: str, o_content: str) -> set:
        contents = [s_content, p_content, o_content]

        ret = []
        for i in contents:
            if i in self.thought_keywords:
                ret += self.thought_keywords[i.lower()]

        ret = set(ret)
        return ret

    def retrieve_relevant_events(self, s_content: str, p_content: str, o_content: str) -> set:
        contents = [s_content, p_content, o_content]

        ret = []
        for i in contents:
            if i in self.event_keywords:
                ret += self.event_keywords[i]

        ret = set(ret)
        return ret



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Time    : 2023/12/19
@Author  : mashenquan
@File    : visualize_graph.py
@Desc    : Visualization tool to visualize the class diagrams or sequence diagrams of the graph repository.
"""
from __future__ import annotations

import re
from abc import ABC
from pathlib import Path
from typing import List, Optional

from pydantic import BaseModel, Field

from metagpt.const import AGGREGATION, COMPOSITION, GENERALIZATION
from metagpt.schema import UMLClassView
from metagpt.utils.common import split_namespace
from metagpt.utils.di_graph_repository import DiGraphRepository
from metagpt.utils.graph_repository import GraphKeyword, GraphRepository


class _VisualClassView(BaseModel):
    """Protected class used by VisualGraphRepo internally.

    Attributes:
        package (str): The package associated with the class.
        uml (Optional[UMLClassView]): Optional UMLClassView associated with the class.
        generalizations (List[str]): List of generalizations for the class.
        compositions (List[str]): List of compositions for the class.
        aggregations (List[str]): List of aggregations for the class.
    """

    package: str
    uml: Optional[UMLClassView] = None
    generalizations: List[str] = Field(default_factory=list)
    compositions: List[str] = Field(default_factory=list)
    aggregations: List[str] = Field(default_factory=list)

    def get_mermaid(self, align: int = 1) -> str:
        """Creates a Markdown Mermaid class diagram text.

        Args:
            align (int): Indent count used for alignment.

        Returns:
            str: The Markdown text representing the Mermaid class diagram.
        """
        if not self.uml:
            return ""
        prefix = "\t" * align

        mermaid_txt = self.uml.get_mermaid(align=align)
        for i in self.generalizations:
            mermaid_txt += f"{prefix}{i} <|-- {self.name}\n"
        for i in self.compositions:
            mermaid_txt += f"{prefix}{i} *-- {self.name}\n"
        for i in self.aggregations:
            mermaid_txt += f"{prefix}{i} o-- {self.name}\n"
        return mermaid_txt

    @property
    def name(self) -> str:
        """Returns the class name without the namespace prefix."""
        return split_namespace(self.package)[-1]


class VisualGraphRepo(ABC):
    """Abstract base class for VisualGraphRepo."""

    graph_db: GraphRepository

    def __init__(self, graph_db):
        self.graph_db = graph_db


class VisualDiGraphRepo(VisualGraphRepo):
    """Implementation of VisualGraphRepo for DiGraph graph repository.

    This class extends VisualGraphRepo to provide specific functionality for a graph repository using DiGraph.
    """

    @classmethod
    async def load_from(cls, filename: str | Path):
        """Load a VisualDiGraphRepo instance from a file."""
        graph_db = await DiGraphRepository.load_from(str(filename))
        return cls(graph_db=graph_db)

    async def get_mermaid_class_view(self) -> str:
        """
        Returns a Markdown Mermaid class diagram code block object.
        """
        rows = await self.graph_db.select(predicate=GraphKeyword.IS, object_=GraphKeyword.CLASS)
        mermaid_txt = "classDiagram\n"
        for r in rows:
            v = await self._get_class_view(ns_class_name=r.subject)
            mermaid_txt += v.get_mermaid()
        return mermaid_txt

    async def _get_class_view(self, ns_class_name: str) -> _VisualClassView:
        """Returns the Markdown Mermaid class diagram code block object for the specified class."""
        rows = await self.graph_db.select(subject=ns_class_name)
        class_view = _VisualClassView(package=ns_class_name)
        for r in rows:
            if r.predicate == GraphKeyword.HAS_CLASS_VIEW:
                class_view.uml = UMLClassView.model_validate_json(r.object_)
            elif r.predicate == GraphKeyword.IS + GENERALIZATION + GraphKeyword.OF:
                name = split_namespace(r.object_)[-1]
                name = self._refine_name(name)
                if name:
                    class_view.generalizations.append(name)
            elif r.predicate == GraphKeyword.IS + COMPOSITION + GraphKeyword.OF:
                name = split_namespace(r.object_)[-1]
                name = self._refine_name(name)
                if name:
                    class_view.compositions.append(name)
            elif r.predicate == GraphKeyword.IS + AGGREGATION + GraphKeyword.OF:
                name = split_namespace(r.object_)[-1]
                name = self._refine_name(name)
                if name:
                    class_view.aggregations.append(name)
        return class_view

    async def get_mermaid_sequence_views(self) -> List[(str, str)]:
        """Returns all Markdown sequence diagrams with their corresponding graph repository keys."""
        sequence_views = []
        rows = await self.graph_db.select(predicate=GraphKeyword.HAS_SEQUENCE_VIEW)
        for r in rows:
            sequence_views.append((r.subject, r.object_))
        return sequence_views

    @staticmethod
    def _refine_name(name: str) -> str:
        """Removes impurity content from the given name.

        Example:
            >>> _refine_name("int")
            ""

            >>> _refine_name('"Class1"')
            'Class1'

            >>> _refine_name("pkg.Class1")
            "Class1"
        """
        name = re.sub(r'^[\'"\\\(\)]+|[\'"\\\(\)]+$', "", name)
        if name in ["int", "float", "bool", "str", "list", "tuple", "set", "dict", "None"]:
            return ""
        if "." in name:
            name = name.split(".")[-1]

        return name

    async def get_mermaid_sequence_view_versions(self) -> List[(str, str)]:
        """Returns all versioned Markdown sequence diagrams with their corresponding graph repository keys."""
        sequence_views = []
        rows = await self.graph_db.select(predicate=GraphKeyword.HAS_SEQUENCE_VIEW_VER)
        for r in rows:
            sequence_views.append((r.subject, r.object_))
        return sequence_views



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Time    : 2023/12/19
@Author  : mashenquan
@File    : rebuild_class_view.py
@Desc    : Reconstructs class diagram from a source code project.
    Implement RFC197, https://deepwisdom.feishu.cn/wiki/VyK0wfq56ivuvjklMKJcmHQknGt
"""

from pathlib import Path
from typing import Optional, Set, Tuple

import aiofiles

from metagpt.actions import Action
from metagpt.config2 import config
from metagpt.const import (
    AGGREGATION,
    COMPOSITION,
    DATA_API_DESIGN_FILE_REPO,
    GENERALIZATION,
    GRAPH_REPO_FILE_REPO,
)
from metagpt.logs import logger
from metagpt.repo_parser import DotClassInfo, RepoParser
from metagpt.schema import UMLClassView
from metagpt.utils.common import concat_namespace, split_namespace
from metagpt.utils.di_graph_repository import DiGraphRepository
from metagpt.utils.graph_repository import GraphKeyword, GraphRepository


class RebuildClassView(Action):
    """
    Reconstructs a graph repository about class diagram from a source code project.

    Attributes:
        graph_db (Optional[GraphRepository]): The optional graph repository.
    """

    graph_db: Optional[GraphRepository] = None

    async def run(self, with_messages=None, format=config.prompt_schema):
        """
        Implementation of `Action`'s `run` method.

        Args:
            with_messages (Optional[Type]): An optional argument specifying messages to react to.
            format (str): The format for the prompt schema.
        """
        graph_repo_pathname = self.context.git_repo.workdir / GRAPH_REPO_FILE_REPO / self.context.git_repo.workdir.name
        self.graph_db = await DiGraphRepository.load_from(str(graph_repo_pathname.with_suffix(".json")))
        repo_parser = RepoParser(base_directory=Path(self.i_context))
        # use pylint
        class_views, relationship_views, package_root = await repo_parser.rebuild_class_views(path=Path(self.i_context))
        await GraphRepository.update_graph_db_with_class_views(self.graph_db, class_views)
        await GraphRepository.update_graph_db_with_class_relationship_views(self.graph_db, relationship_views)
        await GraphRepository.rebuild_composition_relationship(self.graph_db)
        # use ast
        direction, diff_path = self._diff_path(path_root=Path(self.i_context).resolve(), package_root=package_root)
        symbols = repo_parser.generate_symbols()
        for file_info in symbols:
            # Align to the same root directory in accordance with `class_views`.
            file_info.file = self._align_root(file_info.file, direction, diff_path)
            await GraphRepository.update_graph_db_with_file_info(self.graph_db, file_info)
        await self._create_mermaid_class_views()
        await self.graph_db.save()

    async def _create_mermaid_class_views(self) -> str:
        """Creates a Mermaid class diagram using data from the `graph_db` graph repository.

        This method utilizes information stored in the graph repository to generate a Mermaid class diagram.
        Returns:
            mermaid class diagram file name.
        """
        path = self.context.git_repo.workdir / DATA_API_DESIGN_FILE_REPO
        path.mkdir(parents=True, exist_ok=True)
        pathname = path / self.context.git_repo.workdir.name
        filename = str(pathname.with_suffix(".class_diagram.mmd"))
        async with aiofiles.open(filename, mode="w", encoding="utf-8") as writer:
            content = "classDiagram\n"
            logger.debug(content)
            await writer.write(content)
            # class names
            rows = await self.graph_db.select(predicate=GraphKeyword.IS, object_=GraphKeyword.CLASS)
            class_distinct = set()
            relationship_distinct = set()
            for r in rows:
                content = await self._create_mermaid_class(r.subject)
                if content:
                    await writer.write(content)
                    class_distinct.add(r.subject)
            for r in rows:
                content, distinct = await self._create_mermaid_relationship(r.subject)
                if content:
                    logger.debug(content)
                    await writer.write(content)
                    relationship_distinct.update(distinct)
        logger.info(f"classes: {len(class_distinct)}, relationship: {len(relationship_distinct)}")

        if self.i_context:
            r_filename = Path(filename).relative_to(self.context.git_repo.workdir)
            await self.graph_db.insert(
                subject=self.i_context, predicate="hasMermaidClassDiagramFile", object_=str(r_filename)
            )
            logger.info(f"{self.i_context} hasMermaidClassDiagramFile {filename}")
        return filename

    async def _create_mermaid_class(self, ns_class_name) -> str:
        """Generates a Mermaid class diagram for a specific class using data from the `graph_db` graph repository.

        Args:
            ns_class_name (str): The namespace-prefixed name of the class for which the Mermaid class diagram is to be created.

        Returns:
            str: A Mermaid code block object in markdown representing the class diagram.
        """
        fields = split_namespace(ns_class_name)
        if len(fields) > 2:
            # Ignore sub-class
            return ""

        rows = await self.graph_db.select(subject=ns_class_name, predicate=GraphKeyword.HAS_DETAIL)
        if not rows:
            return ""
        dot_class_info = DotClassInfo.model_validate_json(rows[0].object_)
        class_view = UMLClassView.load_dot_class_info(dot_class_info)

        # update uml view
        await self.graph_db.insert(ns_class_name, GraphKeyword.HAS_CLASS_VIEW, class_view.model_dump_json())
        # update uml isCompositeOf
        for c in dot_class_info.compositions:
            await self.graph_db.insert(
                subject=ns_class_name,
                predicate=GraphKeyword.IS + COMPOSITION + GraphKeyword.OF,
                object_=concat_namespace("?", c),
            )

        # update uml isAggregateOf
        for a in dot_class_info.aggregations:
            await self.graph_db.insert(
                subject=ns_class_name,
                predicate=GraphKeyword.IS + AGGREGATION + GraphKeyword.OF,
                object_=concat_namespace("?", a),
            )

        content = class_view.get_mermaid(align=1)
        logger.debug(content)
        return content

    async def _create_mermaid_relationship(self, ns_class_name: str) -> Tuple[Optional[str], Optional[Set]]:
        """Generates a Mermaid class relationship diagram for a specific class using data from the `graph_db` graph repository.

        Args:
            ns_class_name (str): The namespace-prefixed class name for which the Mermaid relationship diagram is to be created.

        Returns:
            Tuple[str, Set]: A tuple containing the relationship diagram as a string and a set of deduplication.
        """
        s_fields = split_namespace(ns_class_name)
        if len(s_fields) > 2:
            # Ignore sub-class
            return None, None

        predicates = {GraphKeyword.IS + v + GraphKeyword.OF: v for v in [GENERALIZATION, COMPOSITION, AGGREGATION]}
        mappings = {
            GENERALIZATION: " <|-- ",
            COMPOSITION: " *-- ",
            AGGREGATION: " o-- ",
        }
        content = ""
        distinct = set()
        for p, v in predicates.items():
            rows = await self.graph_db.select(subject=ns_class_name, predicate=p)
            for r in rows:
                o_fields = split_namespace(r.object_)
                if len(o_fields) > 2:
                    # Ignore sub-class
                    continue
                relationship = mappings.get(v, " .. ")
                link = f"{o_fields[1]}{relationship}{s_fields[1]}"
                distinct.add(link)
                content += f"\t{link}\n"

        return content, distinct

    @staticmethod
    def _diff_path(path_root: Path, package_root: Path) -> (str, str):
        """Returns the difference between the root path and the path information represented in the package name.

        Args:
            path_root (Path): The root path.
            package_root (Path): The package root path.

        Returns:
            Tuple[str, str]: A tuple containing the representation of the difference ("+", "-", "=") and the path detail of the differing part.

        Example:
            >>> _diff_path(path_root=Path("/Users/x/github/MetaGPT"), package_root=Path("/Users/x/github/MetaGPT/metagpt"))
            "-", "metagpt"

            >>> _diff_path(path_root=Path("/Users/x/github/MetaGPT/metagpt"), package_root=Path("/Users/x/github/MetaGPT/metagpt"))
            "=", "."
        """
        if len(str(path_root)) > len(str(package_root)):
            return "+", str(path_root.relative_to(package_root))
        if len(str(path_root)) < len(str(package_root)):
            return "-", str(package_root.relative_to(path_root))
        return "=", "."

    @staticmethod
    def _align_root(path: str, direction: str, diff_path: str) -> str:
        """Aligns the path to the same root represented by `diff_path`.

        Args:
            path (str): The path to be aligned.
            direction (str): The direction of alignment ('+', '-', '=').
            diff_path (str): The path representing the difference.

        Returns:
            str: The aligned path.

        Example:
            >>> _align_root(path="metagpt/software_company.py", direction="+", diff_path="MetaGPT")
            "MetaGPT/metagpt/software_company.py"

            >>> _align_root(path="metagpt/software_company.py", direction="-", diff_path="metagpt")
            "software_company.py"
        """
        if direction == "=":
            return path
        if direction == "+":
            return diff_path + "/" + path
        else:
            return path[len(diff_path) + 1 :]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Time    : 2024/3/11 19:16
@Author  : alexanderwu
@File    : startup.py
"""

# DEPRECATED: This file is deprecated and will be removed in the future.
# The startup.py implementation has been moved to software_company.py



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Desc   : wake_up


from metagpt.ext.stanford_town.actions.st_action import STAction
from metagpt.logs import logger


class WakeUp(STAction):
    name: str = "WakeUp"

    def _func_validate(self, llm_resp: str, prompt: str = None) -> bool:
        try:
            self._func_cleanup(llm_resp, prompt="")
        except Exception:
            return False
        return True

    def _func_cleanup(self, llm_resp: str, prompt: str) -> int:
        cr = int(llm_resp.strip().lower().split("am")[0])
        return cr

    def _func_fail_default_resp(self) -> int:
        fs = 8
        return fs

    async def run(self, role: "STRole"):
        def create_prompt_input(role):
            prompt_input = [
                role.scratch.get_str_iss(),
                role.scratch.get_str_lifestyle(),
                role.scratch.get_str_firstname(),
            ]
            return prompt_input

        prompt_input = create_prompt_input(role)
        prompt = self.generate_prompt_with_tmpl_filename(prompt_input, "wake_up_hour_v1.txt")
        self.fail_default_resp = self._func_fail_default_resp()
        output = await self._run_gpt35_max_tokens(prompt, max_tokens=5)
        logger.info(f"Role: {role.name} Action: {self.cls_name} output: {output}")
        return output



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Time    : 2024/1/4 20:58
@Author  : alexanderwu
@File    : embedding.py
"""
from llama_index.embeddings.openai import OpenAIEmbedding

from metagpt.config2 import config


def get_embedding() -> OpenAIEmbedding:
    llm = config.get_openai_llm()
    if llm is None:
        raise ValueError("To use OpenAIEmbedding, please ensure that config.llm.api_type is correctly set to 'openai'.")

    embedding = OpenAIEmbedding(api_key=llm.api_key, api_base=llm.base_url)
    return embedding



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Desc   : the ActionNode to parse record

from metagpt.actions.action_node import ActionNode

OBSERVATION = ActionNode(
    key="Observation",
    expected_type=str,
    instruction="Provide a description of your observations of the two images. "
    "Subsequently, delineate the distinctions between the first image and the second one.",
    example="",
)

THOUGHT = ActionNode(
    key="Thought",
    expected_type=str,
    instruction="Consider the impact of Action acting on UI elements.",
    example="",
)

DESCRIPTION = ActionNode(
    key="Description",
    expected_type=str,
    instruction="Describe the functionality of the UI element concisely in one or two sentences Do not include "
    "the numeric tag in your description",
    example="",
)

NODES = [OBSERVATION, THOUGHT, DESCRIPTION]

RECORD_PARSE_NODE = ActionNode.from_children("RecordParse", NODES)



from metagpt.environment.werewolf.const import RoleType
from metagpt.ext.werewolf.actions import Impersonate, Speak
from metagpt.ext.werewolf.roles.base_player import BasePlayer


class Werewolf(BasePlayer):
    name: str = RoleType.WEREWOLF.value
    profile: str = RoleType.WEREWOLF.value
    special_action_names: list[str] = ["Hunt"]

    async def _think(self):
        """狼人白天发言时需要伪装，与其他角色不同，因此需要重写_think"""
        await super()._think()
        if isinstance(self.rc.todo, Speak):
            self.rc.todo = Impersonate()
        return True



"""RAG schemas."""
from enum import Enum
from pathlib import Path
from typing import Any, ClassVar, List, Literal, Optional, Union

from chromadb.api.types import CollectionMetadata
from llama_index.core.embeddings import BaseEmbedding
from llama_index.core.indices.base import BaseIndex
from llama_index.core.schema import TextNode
from llama_index.core.vector_stores.types import VectorStoreQueryMode
from pydantic import BaseModel, ConfigDict, Field, PrivateAttr, model_validator

from metagpt.config2 import config
from metagpt.configs.embedding_config import EmbeddingType
from metagpt.logs import logger
from metagpt.rag.interface import RAGObject


class BaseRetrieverConfig(BaseModel):
    """Common config for retrievers.

    If add new subconfig, it is necessary to add the corresponding instance implementation in rag.factories.retriever.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)
    similarity_top_k: int = Field(default=5, description="Number of top-k similar results to return during retrieval.")


class IndexRetrieverConfig(BaseRetrieverConfig):
    """Config for Index-basd retrievers."""

    index: BaseIndex = Field(default=None, description="Index for retriver.")


class FAISSRetrieverConfig(IndexRetrieverConfig):
    """Config for FAISS-based retrievers."""

    dimensions: int = Field(default=0, description="Dimensionality of the vectors for FAISS index construction.")

    _embedding_type_to_dimensions: ClassVar[dict[EmbeddingType, int]] = {
        EmbeddingType.GEMINI: 768,
        EmbeddingType.OLLAMA: 4096,
    }

    @model_validator(mode="after")
    def check_dimensions(self):
        if self.dimensions == 0:
            self.dimensions = config.embedding.dimensions or self._embedding_type_to_dimensions.get(
                config.embedding.api_type, 1536
            )
            if not config.embedding.dimensions and config.embedding.api_type not in self._embedding_type_to_dimensions:
                logger.warning(
                    f"You didn't set dimensions in config when using {config.embedding.api_type}, default to 1536"
                )

        return self


class BM25RetrieverConfig(IndexRetrieverConfig):
    """Config for BM25-based retrievers."""

    _no_embedding: bool = PrivateAttr(default=True)


class ChromaRetrieverConfig(IndexRetrieverConfig):
    """Config for Chroma-based retrievers."""

    persist_path: Union[str, Path] = Field(default="./chroma_db", description="The directory to save data.")
    collection_name: str = Field(default="metagpt", description="The name of the collection.")
    metadata: Optional[CollectionMetadata] = Field(
        default=None, description="Optional metadata to associate with the collection"
    )


class ElasticsearchStoreConfig(BaseModel):
    index_name: str = Field(default="metagpt", description="Name of the Elasticsearch index.")
    es_url: str = Field(default=None, description="Elasticsearch URL.")
    es_cloud_id: str = Field(default=None, description="Elasticsearch cloud ID.")
    es_api_key: str = Field(default=None, description="Elasticsearch API key.")
    es_user: str = Field(default=None, description="Elasticsearch username.")
    es_password: str = Field(default=None, description="Elasticsearch password.")
    batch_size: int = Field(default=200, description="Batch size for bulk indexing.")
    distance_strategy: str = Field(default="COSINE", description="Distance strategy to use for similarity search.")


class ElasticsearchRetrieverConfig(IndexRetrieverConfig):
    """Config for Elasticsearch-based retrievers. Support both vector and text."""

    store_config: ElasticsearchStoreConfig = Field(..., description="ElasticsearchStore config.")
    vector_store_query_mode: VectorStoreQueryMode = Field(
        default=VectorStoreQueryMode.DEFAULT, description="default is vector query."
    )


class ElasticsearchKeywordRetrieverConfig(ElasticsearchRetrieverConfig):
    """Config for Elasticsearch-based retrievers. Support text only."""

    _no_embedding: bool = PrivateAttr(default=True)
    vector_store_query_mode: Literal[VectorStoreQueryMode.TEXT_SEARCH] = Field(
        default=VectorStoreQueryMode.TEXT_SEARCH, description="text query only."
    )


class BaseRankerConfig(BaseModel):
    """Common config for rankers.

    If add new subconfig, it is necessary to add the corresponding instance implementation in rag.factories.ranker.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)
    top_n: int = Field(default=5, description="The number of top results to return.")


class LLMRankerConfig(BaseRankerConfig):
    """Config for LLM-based rankers."""

    llm: Any = Field(
        default=None,
        description="The LLM to rerank with. using Any instead of LLM, as llama_index.core.llms.LLM is pydantic.v1.",
    )


class ColbertRerankConfig(BaseRankerConfig):
    model: str = Field(default="colbert-ir/colbertv2.0", description="Colbert model name.")
    device: str = Field(default="cpu", description="Device to use for sentence transformer.")
    keep_retrieval_score: bool = Field(default=False, description="Whether to keep the retrieval score in metadata.")


class CohereRerankConfig(BaseRankerConfig):
    model: str = Field(default="rerank-english-v3.0")
    api_key: str = Field(default="YOUR_COHERE_API")


class BGERerankConfig(BaseRankerConfig):
    model: str = Field(default="BAAI/bge-reranker-large", description="BAAI Reranker model name.")
    use_fp16: bool = Field(default=True, description="Whether to use fp16 for inference.")


class ObjectRankerConfig(BaseRankerConfig):
    field_name: str = Field(..., description="field name of the object, field's value must can be compared.")
    order: Literal["desc", "asc"] = Field(default="desc", description="the direction of order.")


class BaseIndexConfig(BaseModel):
    """Common config for index.

    If add new subconfig, it is necessary to add the corresponding instance implementation in rag.factories.index.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)
    persist_path: Union[str, Path] = Field(description="The directory of saved data.")


class VectorIndexConfig(BaseIndexConfig):
    """Config for vector-based index."""

    embed_model: BaseEmbedding = Field(default=None, description="Embed model.")


class FAISSIndexConfig(VectorIndexConfig):
    """Config for faiss-based index."""


class ChromaIndexConfig(VectorIndexConfig):
    """Config for chroma-based index."""

    collection_name: str = Field(default="metagpt", description="The name of the collection.")
    metadata: Optional[CollectionMetadata] = Field(
        default=None, description="Optional metadata to associate with the collection"
    )


class BM25IndexConfig(BaseIndexConfig):
    """Config for bm25-based index."""

    _no_embedding: bool = PrivateAttr(default=True)


class ElasticsearchIndexConfig(VectorIndexConfig):
    """Config for es-based index."""

    store_config: ElasticsearchStoreConfig = Field(..., description="ElasticsearchStore config.")
    persist_path: Union[str, Path] = ""


class ElasticsearchKeywordIndexConfig(ElasticsearchIndexConfig):
    """Config for es-based index. no embedding."""

    _no_embedding: bool = PrivateAttr(default=True)


class ObjectNodeMetadata(BaseModel):
    """Metadata of ObjectNode."""

    is_obj: bool = Field(default=True)
    obj: Any = Field(default=None, description="When rag retrieve, will reconstruct obj from obj_json")
    obj_json: str = Field(..., description="The json of object, e.g. obj.model_dump_json()")
    obj_cls_name: str = Field(..., description="The class name of object, e.g. obj.__class__.__name__")
    obj_mod_name: str = Field(..., description="The module name of class, e.g. obj.__class__.__module__")


class ObjectNode(TextNode):
    """RAG add object."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.excluded_llm_metadata_keys = list(ObjectNodeMetadata.model_fields.keys())
        self.excluded_embed_metadata_keys = self.excluded_llm_metadata_keys

    @staticmethod
    def get_obj_metadata(obj: RAGObject) -> dict:
        metadata = ObjectNodeMetadata(
            obj_json=obj.model_dump_json(), obj_cls_name=obj.__class__.__name__, obj_mod_name=obj.__class__.__module__
        )

        return metadata.model_dump()


class OmniParseType(str, Enum):
    """OmniParseType"""

    PDF = "PDF"
    DOCUMENT = "DOCUMENT"


class ParseResultType(str, Enum):
    """The result type for the parser."""

    TXT = "text"
    MD = "markdown"
    JSON = "json"


class OmniParseOptions(BaseModel):
    """OmniParse Options config"""

    result_type: ParseResultType = Field(default=ParseResultType.MD, description="OmniParse result_type")
    parse_type: OmniParseType = Field(default=OmniParseType.DOCUMENT, description="OmniParse parse_type")
    max_timeout: Optional[int] = Field(default=120, description="Maximum timeout for OmniParse service requests")
    num_workers: int = Field(
        default=5,
        gt=0,
        lt=10,
        description="Number of concurrent requests for multiple files",
    )


class OminParseImage(BaseModel):
    image: str = Field(default="", description="image str bytes")
    image_name: str = Field(default="", description="image name")
    image_info: Optional[dict] = Field(default={}, description="image info")


class OmniParsedResult(BaseModel):
    markdown: str = Field(default="", description="markdown text")
    text: str = Field(default="", description="plain text")
    images: Optional[List[OminParseImage]] = Field(default=[], description="images")
    metadata: Optional[dict] = Field(default={}, description="metadata")

    @model_validator(mode="before")
    def set_markdown(cls, values):
        if not values.get("markdown"):
            values["markdown"] = values.get("text")
        return values
